/***************************************************************************
           Fxp.h  - SDL-based fixed point macros, types, and classes.
                        -------------------
    begin                : Mon Mar 3 2003
    copyright            : (C) 2003 by Tyler Montbriand
    email                : tsm@accesscomm.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef __FXP_H__
#define __FXP_H__

#include <SDL/SDL_types.h>
/* Still needed for vector magnitudes, not much else */
#include <math.h>

#include <SDL/begin_code.h>

#ifdef __cplusplus
extern "C" {
#endif

/*! \brief Hack to get 64-bit types working in KDevelop */
#ifdef __GNUC__
#undef Sint64
#define Sint64 long long

#undef Uint64
#define Uint64 unsigned long long
#endif/*__GNUC__*/

/*! \brief Defines the number of bits precision of fixed point #s */
#define SFXP_BITS       10      /*10 bits precision, ~1/100th accuracy */
/*! \brief 2^SFXP_BITS */
#define SFXP_SCALAR     (1<<SFXP_BITS)
/*! \brief Definition for the signed fixed-point type */
typedef Sint32          Sfxp;
/*! \brief The floating point type used converting to/from floating point */
typedef float FxpFloat;

/*! \brief Convert from integer to signed fixed point */
#define TO_SFXP(x)      (((Sint32)(x))<<SFXP_BITS)

/*! \brief Convert from signed fixed point to integer */
#define FROM_SFXP(x)    ((Sint32)((x+((SFXP_SCALAR>>1)))>>SFXP_BITS))

/*! \brief Convert from float to signed fixed point */
#define F_TO_SFXP(x)    ((Sfxp)((x)*SFXP_SCALAR))

/*! \brief Convert from signed fixed point to floating point */
#define SFXP_TO_F(x)    ( ((FxpFloat)(x)) / ((FxpFloat)SFXP_SCALAR) )

/*! \brief Multiplies two signed fixed point numbers */
#define MUL_SFXP(x,y)  FROM_SFXP(((Sint64)(x))*((Sint64)(y)))

/*! \brief Divides two signed fixed point numbers, x/y */
#define DIV_SFXP(x,y)  ((Sfxp)((TO_SFXP((Sint64)(x)))/y))

/*! \brief # of bits in angle type                      */
#define AFXP_BITSIZE  (sizeof(Afxp)*8)
/*! \brief Maximum value of angle type, 2^AFXP_BITSIZE  */
#define AFXP_MAX      (1<<AFXP_BITSIZE)

/*! \brief Defines the bits precision in a fixed-point angle.   */
#define AFXP_BITS       8       /*16-bit angles, 8-bit decimals */
/*! \brief Type definition for the fixed-point angular type.    */
typedef Uint16          Afxp;   /* Angles are a 16-bit type     */
/*! \brief Any more digits than this would be superfluous       */
#define FXP_PI          3.14159f
/*! \brief Convert degrees to fixed-point angles        */
#define DEG_TO_AFXP(x)  ((Afxp)((x)*( ((double)AFXP_MAX) /360.)))
/*! \brief Convert radians to fixed point angles        */
#define RAD_TO_AFXP(x)  DEG_TO_AFXP((x)*(180./3.14159))
/*! \brief Convert fixed point angles to degrees        */
#define AFXP_TO_DEG(x)  ((x)*(360./AFXP_MAX))

/*! \brief Initializes fixed-point trig tables    */

DECLSPEC void SDLCALL FXP_Init();

/*! \brief Returns fixed-point sin value          */
DECLSPEC Sfxp SDLCALL FXP_Sin(Afxp a);
/*! \brief Returns fixed-point inverse-sin value  */
DECLSPEC Afxp SDLCALL FXP_Asin(Sfxp val);
/*! \brief Returns fixed-point cos value          */
DECLSPEC Sfxp SDLCALL FXP_Cos(Afxp a);
/*! \brief Returns fixed-point inverse-cos value  */
DECLSPEC Afxp SDLCALL FXP_Acos(Sfxp val);
/*! \brief Returns fixed-point tan value          */
DECLSPEC Sfxp SDLCALL FXP_Tan(Afxp a);
/*! \brief Returns fixed-point inverse-tan value  */
DECLSPEC Afxp SDLCALL FXP_Atan(Sfxp val);
/*! \brief Returns the trajectory of the angle of the specified vector */
DECLSPEC Afxp SDLCALL FXP_Traj(Sfxp y, Sfxp x);


#if !defined(__cplusplus)
#  if !(defined(WIN32)||defined(WIN32_WCE))
/* #warn directives don't work in MSVC */
#    warning "Warning - c++ compiler not detected, excluding class definitions"
#  endif
#else

/*! \brief Class used to handle fixed-point x/y coordinates */
class DECLSPEC SfxpCoord
{
public:
  /*! \brief Floating point constructor */
  inline SfxpCoord(FxpFloat xin, FxpFloat yin)
  { SetFloat(xin,yin);  }
  
  /*! \brief Integer constructor */
  inline SfxpCoord(int xin=0,int yin=0)
  { SetInt(xin,yin);  }

  /*! \brief Comparison operator */
  inline bool operator ==(SfxpCoord &coord)
  { return((x=coord.x)&&(y==coord.y));  }

  /*! \brief Comparison operator */
  inline bool operator !=(SfxpCoord &coord)
  { return(!((*this)==coord));  }

  /*! \brief Vector addition operator. */  
  inline SfxpCoord operator +(SfxpCoord coord)
  {
    SfxpCoord ncoord;

    ncoord.SetFxp(x+coord.x,y+coord.y);
    return(ncoord);
  }
  
  /*! \brief Vector subtraction operator. */
  inline SfxpCoord operator -(SfxpCoord coord)
  {
    SfxpCoord ncoord;

    ncoord.SetFxp(x-coord.x,y-coord.y);
    return(ncoord);
  }

  /*! \brief Adds passed vector to stored vector. */
  inline SfxpCoord operator +=(SfxpCoord coord)
  {
    x+=coord.x;
    y+=coord.y;
    return(*this);
  }

  /*! \brief Multiplies the vector by a signed fixed point scalar value. */
  inline SfxpCoord FxpScalar(Sfxp val)
  {
    x=MUL_SFXP(x,val);
    y=MUL_SFXP(y,val);
    return(*this);
  }

  inline Afxp Trajectory()
  {
    return(FXP_Traj(y,x));
  }

  /*! \brief Multiplies the vector by a signed fixed-point integer value. */
  inline SfxpCoord IntScalar(int val)
  {
    x*=val;
    y*=val;
    return(*this);
  }

  /*! \brief Returns true if the truncated integer portions match. */
  inline bool MajorDigitsEqual(SfxpCoord coord) const
  {
    if((X()==coord.X())&&(Y()==coord.Y()))  return(true);
    else                                    return(false);
  }

  /*! \brief Returns the truncated integer portion of the X coordinate. */
  inline int X() const
  { return(FROM_SFXP(x)); }

  /*! \brief Returns the truncated integer portion of the Y coordinate. */
  inline int Y() const
  { return(FROM_SFXP(y)); }

  /*! \brief Returns the raw fixed-point X coordinate. */
  inline int Xfxp() const
  { return(x);  }

  /*! \brief Returns the raw fixed-point Y coordinate. */  
  inline int Yfxp() const
  { return(y);  }

  /*! \brief Returns the X coordinate, converted to floating point */
  inline FxpFloat Xfloat()
  { return((FxpFloat)x/(FxpFloat)SFXP_SCALAR);  }

  /*! \brief Returns the Y coordinate, converted to floating point */  
  inline FxpFloat Yfloat()
  { return((FxpFloat)y/(FxpFloat)SFXP_SCALAR);  }

  /*! \brief Sets X and Y coordinates, converting from floating point */
  inline SfxpCoord SetFloat(FxpFloat xin, FxpFloat yin)
  {
    Sfxp xsf,ysf;

    xsf=(Sfxp)(xin*(FxpFloat)SFXP_SCALAR);
    ysf=(Sfxp)(yin*(FxpFloat)SFXP_SCALAR);
    SetFxp(xsf,ysf);

    return(*this);
  }

  /*! \brief Sets X and Y coordinates from raw fixed point values */
  inline SfxpCoord SetFxp(Sfxp xin, Sfxp yin)
  {
    x=xin;
    y=yin;
    return(*this);
  }

  /*! \brief Initializes coordinates, converting from integer values */
  inline SfxpCoord SetInt(Sint32 xin, Sint32 yin)
  {
    x=TO_SFXP(xin);
    y=TO_SFXP(yin);
    return(*this);
  }  

  /*! \brief Returns zero if the vector is zero in magnitude */
  inline int Nonzero()
  { return(x|y);  }

  /*! \brief Initializes coordinates to the specified angle and magnitude */
  inline void Vector(Sfxp magnitude, Afxp angle)
  {
    x=MUL_SFXP(FXP_Cos(angle),magnitude);
    y=MUL_SFXP(FXP_Sin(angle),magnitude);
  }

/*  inline void Print(FILE *fp=stdout)
  {
    fprintf(fp,"(%+.3f,%+.3f)\n",Xfloat(),Yfloat());
  }*/

  Sfxp Magnitude()
  {
    if(Nonzero())
    {
      FxpFloat x2=SFXP_TO_F(x)*SFXP_TO_F(x);
      FxpFloat y2=SFXP_TO_F(y)*SFXP_TO_F(y);

      return(F_TO_SFXP(sqrt(x2+y2)));
    }
    else
    {
      return(0);
    }
  }

private:
  /*! \brief Actual signed fixed-point coordinate values */
  Sfxp x,y;
};

/*! \brief Class used to handle fixed-point angles */
class DECLSPEC Angle
{
public:
  /*! \brief Initializes from a raw fixed-point angle */
  inline Angle(Afxp ain)
  { angle=ain;            }

  /*! \brief Initializes from an integer number of degrees */
  inline Angle(int degrees)
  { SetDegrees(degrees);  }

  /*! \brief Initializes from a floating-point radian value */
  inline Angle(FxpFloat radians)
  { SetRadians(radians);  }

  /*! \brief Converts to an integer number of degrees */
  inline int Degrees() const
  { return((int)AFXP_TO_DEG(angle));  }

  /*! \brief Converts to a floating-point radian value */
  inline FxpFloat Radians() const
  { return((FxpFloat)((FXP_PI/180.)*AFXP_TO_DEG(angle))); }

  /*! \brief Returns the raw floating-point angle */
  inline Afxp Fxp() const
  { return(angle);  }

  /*! \brief Adds specified angle to the stored angle. */
  inline Angle operator+=(Afxp ain)
  {
    angle+=ain;
    return(*this);
  }

  /*! \brief Subtracts specified angle from the stored angle. */
  inline Angle operator-=(Afxp ain)
  {
    angle-=ain;
    return(*this);
  }

  /*! \brief Returns the sum of the two angles. */
  inline Angle operator+(const Angle &ain)
  { return(Angle(angle+ain.Fxp())); }

  /*! \brief Sets to an integer number of degrees. */
  inline Angle SetDegrees(int degrees)
  {
    angle=DEG_TO_AFXP(degrees);
    return(*this);
  }
  /*! \brief Sets to a floating-point radian value */
  inline Angle SetRadians(double radians)
  {
    angle=RAD_TO_AFXP(radians);
    return(*this);
  }

  /*! \brief Returns the sine value for the angle. */
  inline Sfxp Sin() const
  { return(FXP_Sin(angle)); }

  /*! \brief Returns the cosine value for the angle. */
  inline Sfxp Cos() const
  { return(FXP_Cos(angle)); }

  /*! \brief Returns the tan value for the angle. */
  inline Sfxp Tan() const
  { return(FXP_Tan(angle)); }

private:
  /*! \brief The actual fixed-point angle data */
  Afxp angle;
};
#endif/*__cplusplus*/

#ifdef __cplusplus
}
#endif

#include <SDL/close_code.h>

#endif/*__FXP_H__*/
