/***************************************************************************
                          Fxp_trig.c  -  description
                             -------------------
    begin                : Sun Jun 8 2003
    copyright            : (C) 2003 by Tyler Montbriand
    email                : tsm@accesscomm.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdio.h>
#include <math.h>
#include "Fxp.h"

/*! \brief # of elements in trig tables */
#define TABLE_SIZE    2048

#define ACOS_TEST(x)  FXP_Acos(FXP_Cos(x))
#define ASIN_TEST(x)  FXP_Asin(FXP_Sin(x))
#define ATAN_TEST(x)  FXP_Atan(FXP_Tan(x))

/*! \brief Converts a table-index to an angle */
#define TABLEPOS_TO_ANGLE(n)  ((Afxp)((AFXP_MAX*n)/TABLE_SIZE))
/*! \brief Converts an angle to a table pos */
#define ANGLE_TO_TABLEPOS(a)  (((TABLE_SIZE*(int)a)/AFXP_MAX)%TABLE_SIZE)
/*! \brief Returns average of two values */
#define AVERAGE(x,y)          ((x+y)/2)

/*! \brief Table of sin values vs angle */
extern const Sfxp __SIN_TABLE[];
/*! \brief Table of cos values vs angle */
extern const Sfxp __COS_TABLE[];
/*! \brief Table of tan values vs angle */
extern const Sfxp __TAN_TABLE[];

/*! \brief Binary search for specified area, sorted in increasing order */
int FindInTable(Sfxp val, int min, int max, const Sfxp table[TABLE_SIZE]);

/*! \brief Prints contents of trig tables to a C file */
int FXP_DumpTrig(const char *fname)
{
  int n=-1;
  FILE *fp=fopen(fname,"w");
  if(fp==NULL)  return(-1);

  fprintf(fp,"/* Put this portion in the .h file */\n");
  fprintf(fp,"extern Sfxp __SIN_TABLE[];\n");
  fprintf(fp,"extern Sfxp __COS_TABLE[];\n");
  fprintf(fp,"extern Sfxp __TAN_TABLE[];\n\n");

  fprintf(fp,"/* Put this portion in the .c file */\n");
  fprintf(fp,"Sfxp __SIN_TABLE[%d] = {\n",TABLE_SIZE);
  for(n=0; n<TABLE_SIZE; n++)
  {
    if(n==(TABLE_SIZE-1))       fprintf(fp,"%5d\n",__SIN_TABLE[n]);
    else                        fprintf(fp,"%5d, ",__SIN_TABLE[n]);
    if((((n+1)%10)==0)&&(n!=0)) fprintf(fp,"\n");
  }
  fprintf(fp,"};\n");

  fprintf(fp,"Sfxp __COS_TABLE[%d] = {\n",TABLE_SIZE);
  for(n=0; n<TABLE_SIZE; n++)
  {
    if(n==(TABLE_SIZE-1))       fprintf(fp,"%5d\n",__COS_TABLE[n]);
    else                        fprintf(fp,"%5d, ",__COS_TABLE[n]);
    if((((n+1)%10)==0)&&(n!=0)) fprintf(fp,"\n");
  }
  fprintf(fp,"};\n");


  fprintf(fp,"Sfxp __TAN_TABLE[%d] = {\n",TABLE_SIZE);
  for(n=0; n<TABLE_SIZE; n++)
  {
    if(n==(TABLE_SIZE-1))       fprintf(fp,"%10d\n",__TAN_TABLE[n]);
    else                        fprintf(fp,"%10d, ",__TAN_TABLE[n]);
    if((((n+1)%5)==0)&&(n!=0))  fprintf(fp,"\n");
  }
  fprintf(fp,"};\n");


  return(0);
}

/*! \brief Fills the trig tables with values */
void FXP_Init()
{
  int n=-1;

#if 0 /* Not valid when table is read-only  */
  for(n=0; n<TABLE_SIZE; n++)
  {
    Angle a(TABLEPOS_TO_ANGLE(n));
    double rad=a.Radians();

    __SIN_TABLE[n]=F_TO_SFXP(sin(rad));
    __COS_TABLE[n]=F_TO_SFXP(cos(rad));
    __TAN_TABLE[n]=F_TO_SFXP(tan(rad));
  }
#endif


  FXP_DumpTrig("trig.txt");
}

Afxp FXP_Asin(Sfxp val)
{
  int index=-1;
  /* Special case - sin(0) = 0 */
  if(val==0)
  {
    return(0);
  }
  else if(val>0)
  {
    /* Special case: sin(90 degrees) = 1 */
    if(val==TO_SFXP(1))
    {
      return(AFXP_MAX/4);
    }
    else
    {
      index=FindInTable(val,0,(TABLE_SIZE/4)+1,__SIN_TABLE);
    }
  }
  else  /* val < 0 */
  {
    /* Special case: sin(270 deg)== -1 */
    if(val==TO_SFXP(-1))
    {
      return((AFXP_MAX*3)/4);
    }
    else
    {
      index=FindInTable(val,(TABLE_SIZE*3)/4,TABLE_SIZE-1,__SIN_TABLE);
      index-=((TABLE_SIZE*3)/4);
      index=((TABLE_SIZE*3)/4)-index;
    }
  }

  return(TABLEPOS_TO_ANGLE(index));
}

Afxp FXP_Acos(Sfxp val)
{
  int index=-1;


  /*Special case - cos(90 degrees) == 0 */
  if(val==TO_SFXP(0))
  {
    return(AFXP_MAX/4);
  }
  else if(val>0)
  {
    /* Special case -  cos(0) == 1 */
    if(val==TO_SFXP(1))
    {
      return(0);
    }
    else
    {
      index=FindInTable(val,(TABLE_SIZE*3)/4,TABLE_SIZE-1,__COS_TABLE);
      index-=(TABLE_SIZE*3)/4;
      index=(TABLE_SIZE/4)-index;
    }
  }
  else
  {
    index=FindInTable(val,TABLE_SIZE/2,(TABLE_SIZE*3)/4+1,__COS_TABLE);
    index-=TABLE_SIZE/2;
    index=(TABLE_SIZE/2)-index;
  }

  return(TABLEPOS_TO_ANGLE(index));
}

Afxp FXP_Atan(Sfxp val)
{
  int index=-1;

  /* Special case - tan(0) == 0 */
  if(val==0)
  {
    return(0);
  }
  else if(val>0)
  {
    index=FindInTable(val,0,(TABLE_SIZE/4)+1,__TAN_TABLE);
  }
  else
  {
    index=FindInTable(val,(TABLE_SIZE/4),(TABLE_SIZE/2)+1,__TAN_TABLE);
  }

  return(TABLEPOS_TO_ANGLE(index)%DEG_TO_AFXP(90));
}

Sfxp FXP_Sin(Afxp a)
{
  return(__SIN_TABLE[ANGLE_TO_TABLEPOS(a)]);
}

Sfxp FXP_Cos(Afxp a)
{
  return(__COS_TABLE[ANGLE_TO_TABLEPOS(a)]);
}

Sfxp FXP_Tan(Afxp a)
{
  return(__TAN_TABLE[ANGLE_TO_TABLEPOS(a)]);
}

int FindInTable(Sfxp val, int min, int max, const Sfxp table[TABLE_SIZE])
{
  while(min!=max)
  {
    int pos=AVERAGE(min,max);
    int fval=table[pos];

    if(min+1==max)
    {
      return(pos);
    }
    else if(fval<val)  /*Need to look higher */
    {
      min=pos;
    }
    else if(fval>val) /* Need to look lower */
    {
      max=pos;
    }
    else
    {
      return(pos);
    }
  }

  return(min);
}

/**
 *  Unlike Tan, this returns the correct angle in the range from 0 to
 *  360 degrees.
 */
Afxp FXP_Traj(Sfxp y, Sfxp x)
{
  Afxp a=0;
  
  if((x!=0)&&(y!=0))
  {
    a=FXP_Atan(DIV_SFXP(y,x));
    if(x>0)
    {
      if(y<0)
      {
        a+=DEG_TO_AFXP(270);
      }
    }
    else if(x<0)
    {
      if(y>0)
      {
        a+=DEG_TO_AFXP(90);        
      }
      else if(y<0)
      {
        a+=DEG_TO_AFXP(180);
      }
    }
  
  }
  else if(y!=0)
  {
    if(y>0)       a=DEG_TO_AFXP(90);
    else if(y<0)  a=DEG_TO_AFXP(270);
  }
  else if(x!=0)
  {
    if(x<0)       a=DEG_TO_AFXP(180);
    else if(x>0)  a=DEG_TO_AFXP(0);
  }

  return(a);
}


const Sfxp __SIN_TABLE[TABLE_SIZE] = {
    0,     3,     6,     9,    12,    15,    18,    21,    25,    28,
   31,    34,    37,    40,    43,    47,    50,    53,    56,    59,
   62,    65,    69,    72,    75,    78,    81,    84,    87,    90,
   94,    97,   100,   103,   106,   109,   112,   115,   119,   122,
  125,   128,   131,   134,   137,   140,   144,   147,   150,   153,
  156,   159,   162,   165,   168,   171,   175,   178,   181,   184,
  187,   190,   193,   196,   199,   202,   205,   209,   212,   215,
  218,   221,   224,   227,   230,   233,   236,   239,   242,   245,
  248,   251,   254,   257,   260,   264,   267,   270,   273,   276,
  279,   282,   285,   288,   291,   294,   297,   300,   303,   306,
  309,   312,   315,   318,   321,   324,   327,   330,   333,   336,
  339,   342,   344,   347,   350,   353,   356,   359,   362,   365,
  368,   371,   374,   377,   380,   383,   386,   388,   391,   394,
  397,   400,   403,   406,   409,   412,   414,   417,   420,   423,
  426,   429,   432,   434,   437,   440,   443,   446,   449,   451,
  454,   457,   460,   463,   466,   468,   471,   474,   477,   479,
  482,   485,   488,   491,   493,   496,   499,   501,   504,   507,
  510,   512,   515,   518,   521,   523,   526,   529,   531,   534,
  537,   539,   542,   545,   547,   550,   553,   555,   558,   561,
  563,   566,   568,   571,   574,   576,   579,   581,   584,   587,
  589,   592,   594,   597,   599,   602,   604,   607,   609,   612,
  615,   617,   620,   622,   625,   627,   629,   632,   634,   637,
  639,   642,   644,   647,   649,   652,   654,   656,   659,   661,
  664,   666,   668,   671,   673,   675,   678,   680,   683,   685,
  687,   690,   692,   694,   696,   699,   701,   703,   706,   708,
  710,   712,   715,   717,   719,   721,   724,   726,   728,   730,
  732,   735,   737,   739,   741,   743,   745,   748,   750,   752,
  754,   756,   758,   760,   762,   765,   767,   769,   771,   773,
  775,   777,   779,   781,   783,   785,   787,   789,   791,   793,
  795,   797,   799,   801,   803,   805,   807,   809,   811,   813,
  814,   816,   818,   820,   822,   824,   826,   828,   829,   831,
  833,   835,   837,   839,   840,   842,   844,   846,   847,   849,
  851,   853,   854,   856,   858,   860,   861,   863,   865,   866,
  868,   870,   871,   873,   875,   876,   878,   879,   881,   883,
  884,   886,   887,   889,   890,   892,   894,   895,   897,   898,
  900,   901,   903,   904,   906,   907,   908,   910,   911,   913,
  914,   916,   917,   918,   920,   921,   922,   924,   925,   927,
  928,   929,   930,   932,   933,   934,   936,   937,   938,   939,
  941,   942,   943,   944,   946,   947,   948,   949,   950,   951,
  953,   954,   955,   956,   957,   958,   959,   960,   962,   963,
  964,   965,   966,   967,   968,   969,   970,   971,   972,   973,
  974,   975,   976,   977,   978,   978,   979,   980,   981,   982,
  983,   984,   985,   986,   986,   987,   988,   989,   990,   990,
  991,   992,   993,   994,   994,   995,   996,   997,   997,   998,
  999,   999,  1000,  1001,  1001,  1002,  1003,  1003,  1004,  1004,
 1005,  1006,  1006,  1007,  1007,  1008,  1008,  1009,  1009,  1010,
 1010,  1011,  1011,  1012,  1012,  1013,  1013,  1014,  1014,  1015,
 1015,  1015,  1016,  1016,  1017,  1017,  1017,  1018,  1018,  1018,
 1019,  1019,  1019,  1019,  1020,  1020,  1020,  1020,  1021,  1021,
 1021,  1021,  1022,  1022,  1022,  1022,  1022,  1022,  1023,  1023,
 1023,  1023,  1023,  1023,  1023,  1023,  1023,  1023,  1023,  1023,
 1023,  1023,  1023,  1023,  1023,  1023,  1023,  1023,  1023,  1023,
 1023,  1023,  1023,  1023,  1023,  1023,  1023,  1022,  1022,  1022,
 1022,  1022,  1022,  1021,  1021,  1021,  1021,  1020,  1020,  1020,
 1020,  1019,  1019,  1019,  1019,  1018,  1018,  1018,  1017,  1017,
 1017,  1016,  1016,  1015,  1015,  1015,  1014,  1014,  1013,  1013,
 1012,  1012,  1011,  1011,  1010,  1010,  1009,  1009,  1008,  1008,
 1007,  1007,  1006,  1006,  1005,  1004,  1004,  1003,  1003,  1002,
 1001,  1001,  1000,   999,   999,   998,   997,   997,   996,   995,
  994,   994,   993,   992,   991,   990,   990,   989,   988,   987,
  986,   986,   985,   984,   983,   982,   981,   980,   979,   978,
  978,   977,   976,   975,   974,   973,   972,   971,   970,   969,
  968,   967,   966,   965,   964,   963,   962,   960,   959,   958,
  957,   956,   955,   954,   953,   951,   950,   949,   948,   947,
  946,   944,   943,   942,   941,   939,   938,   937,   936,   934,
  933,   932,   930,   929,   928,   927,   925,   924,   922,   921,
  920,   918,   917,   916,   914,   913,   911,   910,   908,   907,
  906,   904,   903,   901,   900,   898,   897,   895,   894,   892,
  890,   889,   887,   886,   884,   883,   881,   879,   878,   876,
  875,   873,   871,   870,   868,   866,   865,   863,   861,   860,
  858,   856,   854,   853,   851,   849,   847,   846,   844,   842,
  840,   839,   837,   835,   833,   831,   829,   828,   826,   824,
  822,   820,   818,   816,   814,   813,   811,   809,   807,   805,
  803,   801,   799,   797,   795,   793,   791,   789,   787,   785,
  783,   781,   779,   777,   775,   773,   771,   769,   767,   765,
  762,   760,   758,   756,   754,   752,   750,   748,   745,   743,
  741,   739,   737,   735,   732,   730,   728,   726,   724,   721,
  719,   717,   715,   712,   710,   708,   706,   703,   701,   699,
  696,   694,   692,   690,   687,   685,   683,   680,   678,   675,
  673,   671,   668,   666,   664,   661,   659,   656,   654,   652,
  649,   647,   644,   642,   639,   637,   634,   632,   629,   627,
  625,   622,   620,   617,   615,   612,   609,   607,   604,   602,
  599,   597,   594,   592,   589,   587,   584,   581,   579,   576,
  574,   571,   568,   566,   563,   561,   558,   555,   553,   550,
  547,   545,   542,   539,   537,   534,   531,   529,   526,   523,
  521,   518,   515,   512,   510,   507,   504,   501,   499,   496,
  493,   491,   488,   485,   482,   479,   477,   474,   471,   468,
  466,   463,   460,   457,   454,   451,   449,   446,   443,   440,
  437,   434,   432,   429,   426,   423,   420,   417,   414,   412,
  409,   406,   403,   400,   397,   394,   391,   388,   386,   383,
  380,   377,   374,   371,   368,   365,   362,   359,   356,   353,
  350,   347,   344,   342,   339,   336,   333,   330,   327,   324,
  321,   318,   315,   312,   309,   306,   303,   300,   297,   294,
  291,   288,   285,   282,   279,   276,   273,   270,   267,   264,
  260,   257,   254,   251,   248,   245,   242,   239,   236,   233,
  230,   227,   224,   221,   218,   215,   212,   209,   205,   202,
  199,   196,   193,   190,   187,   184,   181,   178,   175,   171,
  168,   165,   162,   159,   156,   153,   150,   147,   144,   140,
  137,   134,   131,   128,   125,   122,   119,   115,   112,   109,
  106,   103,   100,    97,    94,    90,    87,    84,    81,    78,
   75,    72,    69,    65,    62,    59,    56,    53,    50,    47,
   43,    40,    37,    34,    31,    28,    25,    21,    18,    15,
   12,     9,     6,     3,     0,    -3,    -6,    -9,   -12,   -15,
  -18,   -21,   -25,   -28,   -31,   -34,   -37,   -40,   -43,   -47,
  -50,   -53,   -56,   -59,   -62,   -65,   -69,   -72,   -75,   -78,
  -81,   -84,   -87,   -90,   -94,   -97,  -100,  -103,  -106,  -109,
 -112,  -115,  -119,  -122,  -125,  -128,  -131,  -134,  -137,  -140,
 -144,  -147,  -150,  -153,  -156,  -159,  -162,  -165,  -168,  -171,
 -175,  -178,  -181,  -184,  -187,  -190,  -193,  -196,  -199,  -202,
 -205,  -209,  -212,  -215,  -218,  -221,  -224,  -227,  -230,  -233,
 -236,  -239,  -242,  -245,  -248,  -251,  -254,  -257,  -260,  -264,
 -267,  -270,  -273,  -276,  -279,  -282,  -285,  -288,  -291,  -294,
 -297,  -300,  -303,  -306,  -309,  -312,  -315,  -318,  -321,  -324,
 -327,  -330,  -333,  -336,  -339,  -342,  -344,  -347,  -350,  -353,
 -356,  -359,  -362,  -365,  -368,  -371,  -374,  -377,  -380,  -383,
 -386,  -388,  -391,  -394,  -397,  -400,  -403,  -406,  -409,  -412,
 -414,  -417,  -420,  -423,  -426,  -429,  -432,  -434,  -437,  -440,
 -443,  -446,  -449,  -451,  -454,  -457,  -460,  -463,  -466,  -468,
 -471,  -474,  -477,  -479,  -482,  -485,  -488,  -490,  -493,  -496,
 -499,  -501,  -504,  -507,  -510,  -512,  -515,  -518,  -521,  -523,
 -526,  -529,  -531,  -534,  -537,  -539,  -542,  -545,  -547,  -550,
 -553,  -555,  -558,  -561,  -563,  -566,  -568,  -571,  -574,  -576,
 -579,  -581,  -584,  -587,  -589,  -592,  -594,  -597,  -599,  -602,
 -604,  -607,  -609,  -612,  -615,  -617,  -620,  -622,  -625,  -627,
 -629,  -632,  -634,  -637,  -639,  -642,  -644,  -647,  -649,  -652,
 -654,  -656,  -659,  -661,  -664,  -666,  -668,  -671,  -673,  -675,
 -678,  -680,  -683,  -685,  -687,  -689,  -692,  -694,  -696,  -699,
 -701,  -703,  -706,  -708,  -710,  -712,  -715,  -717,  -719,  -721,
 -724,  -726,  -728,  -730,  -732,  -735,  -737,  -739,  -741,  -743,
 -745,  -748,  -750,  -752,  -754,  -756,  -758,  -760,  -762,  -765,
 -767,  -769,  -771,  -773,  -775,  -777,  -779,  -781,  -783,  -785,
 -787,  -789,  -791,  -793,  -795,  -797,  -799,  -801,  -803,  -805,
 -807,  -809,  -811,  -813,  -814,  -816,  -818,  -820,  -822,  -824,
 -826,  -828,  -829,  -831,  -833,  -835,  -837,  -839,  -840,  -842,
 -844,  -846,  -847,  -849,  -851,  -853,  -854,  -856,  -858,  -860,
 -861,  -863,  -865,  -866,  -868,  -870,  -871,  -873,  -875,  -876,
 -878,  -879,  -881,  -883,  -884,  -886,  -887,  -889,  -890,  -892,
 -894,  -895,  -897,  -898,  -900,  -901,  -903,  -904,  -906,  -907,
 -908,  -910,  -911,  -913,  -914,  -916,  -917,  -918,  -920,  -921,
 -922,  -924,  -925,  -927,  -928,  -929,  -930,  -932,  -933,  -934,
 -936,  -937,  -938,  -939,  -941,  -942,  -943,  -944,  -946,  -947,
 -948,  -949,  -950,  -951,  -953,  -954,  -955,  -956,  -957,  -958,
 -959,  -960,  -962,  -963,  -964,  -965,  -966,  -967,  -968,  -969,
 -970,  -971,  -972,  -973,  -974,  -975,  -976,  -977,  -978,  -978,
 -979,  -980,  -981,  -982,  -983,  -984,  -985,  -986,  -986,  -987,
 -988,  -989,  -990,  -990,  -991,  -992,  -993,  -994,  -994,  -995,
 -996,  -997,  -997,  -998,  -999,  -999, -1000, -1001, -1001, -1002,
-1003, -1003, -1004, -1004, -1005, -1006, -1006, -1007, -1007, -1008,
-1008, -1009, -1009, -1010, -1010, -1011, -1011, -1012, -1012, -1013,
-1013, -1014, -1014, -1015, -1015, -1015, -1016, -1016, -1017, -1017,
-1017, -1018, -1018, -1018, -1019, -1019, -1019, -1019, -1020, -1020,
-1020, -1020, -1021, -1021, -1021, -1021, -1022, -1022, -1022, -1022,
-1022, -1022, -1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023,
-1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023,
-1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023,
-1023, -1022, -1022, -1022, -1022, -1022, -1022, -1021, -1021, -1021,
-1021, -1020, -1020, -1020, -1020, -1019, -1019, -1019, -1019, -1018,
-1018, -1018, -1017, -1017, -1017, -1016, -1016, -1015, -1015, -1015,
-1014, -1014, -1013, -1013, -1012, -1012, -1011, -1011, -1010, -1010,
-1009, -1009, -1008, -1008, -1007, -1007, -1006, -1006, -1005, -1004,
-1004, -1003, -1003, -1002, -1001, -1001, -1000,  -999,  -999,  -998,
 -997,  -997,  -996,  -995,  -994,  -994,  -993,  -992,  -991,  -990,
 -990,  -989,  -988,  -987,  -986,  -986,  -985,  -984,  -983,  -982,
 -981,  -980,  -979,  -978,  -978,  -977,  -976,  -975,  -974,  -973,
 -972,  -971,  -970,  -969,  -968,  -967,  -966,  -965,  -964,  -963,
 -962,  -960,  -959,  -958,  -957,  -956,  -955,  -954,  -953,  -951,
 -950,  -949,  -948,  -947,  -946,  -944,  -943,  -942,  -941,  -939,
 -938,  -937,  -936,  -934,  -933,  -932,  -930,  -929,  -928,  -927,
 -925,  -924,  -922,  -921,  -920,  -918,  -917,  -916,  -914,  -913,
 -911,  -910,  -908,  -907,  -906,  -904,  -903,  -901,  -900,  -898,
 -897,  -895,  -894,  -892,  -890,  -889,  -887,  -886,  -884,  -883,
 -881,  -879,  -878,  -876,  -875,  -873,  -871,  -870,  -868,  -866,
 -865,  -863,  -861,  -860,  -858,  -856,  -854,  -853,  -851,  -849,
 -847,  -846,  -844,  -842,  -840,  -839,  -837,  -835,  -833,  -831,
 -829,  -828,  -826,  -824,  -822,  -820,  -818,  -816,  -814,  -813,
 -811,  -809,  -807,  -805,  -803,  -801,  -799,  -797,  -795,  -793,
 -791,  -789,  -787,  -785,  -783,  -781,  -779,  -777,  -775,  -773,
 -771,  -769,  -767,  -765,  -762,  -760,  -758,  -756,  -754,  -752,
 -750,  -748,  -745,  -743,  -741,  -739,  -737,  -735,  -732,  -730,
 -728,  -726,  -724,  -721,  -719,  -717,  -715,  -712,  -710,  -708,
 -706,  -703,  -701,  -699,  -696,  -694,  -692,  -690,  -687,  -685,
 -683,  -680,  -678,  -675,  -673,  -671,  -668,  -666,  -664,  -661,
 -659,  -656,  -654,  -652,  -649,  -647,  -644,  -642,  -639,  -637,
 -634,  -632,  -630,  -627,  -625,  -622,  -620,  -617,  -615,  -612,
 -609,  -607,  -604,  -602,  -599,  -597,  -594,  -592,  -589,  -587,
 -584,  -581,  -579,  -576,  -574,  -571,  -568,  -566,  -563,  -561,
 -558,  -555,  -553,  -550,  -547,  -545,  -542,  -539,  -537,  -534,
 -531,  -529,  -526,  -523,  -521,  -518,  -515,  -512,  -510,  -507,
 -504,  -501,  -499,  -496,  -493,  -491,  -488,  -485,  -482,  -479,
 -477,  -474,  -471,  -468,  -466,  -463,  -460,  -457,  -454,  -451,
 -449,  -446,  -443,  -440,  -437,  -434,  -432,  -429,  -426,  -423,
 -420,  -417,  -414,  -412,  -409,  -406,  -403,  -400,  -397,  -394,
 -391,  -388,  -386,  -383,  -380,  -377,  -374,  -371,  -368,  -365,
 -362,  -359,  -356,  -353,  -350,  -347,  -344,  -342,  -339,  -336,
 -333,  -330,  -327,  -324,  -321,  -318,  -315,  -312,  -309,  -306,
 -303,  -300,  -297,  -294,  -291,  -288,  -285,  -282,  -279,  -276,
 -273,  -270,  -267,  -264,  -260,  -257,  -254,  -251,  -248,  -245,
 -242,  -239,  -236,  -233,  -230,  -227,  -224,  -221,  -218,  -215,
 -212,  -209,  -205,  -202,  -199,  -196,  -193,  -190,  -187,  -184,
 -181,  -178,  -175,  -171,  -168,  -165,  -162,  -159,  -156,  -153,
 -150,  -147,  -144,  -140,  -137,  -134,  -131,  -128,  -125,  -122,
 -119,  -115,  -112,  -109,  -106,  -103,  -100,   -97,   -94,   -90,
  -87,   -84,   -81,   -78,   -75,   -72,   -69,   -65,   -62,   -59,
  -56,   -53,   -50,   -47,   -43,   -40,   -37,   -34,   -31,   -28,
  -25,   -21,   -18,   -15,   -12,    -9,    -6,    -3
};

const Sfxp __COS_TABLE[TABLE_SIZE] = {
 1024,  1023,  1023,  1023,  1023,  1023,  1023,  1023,  1023,  1023,
 1023,  1023,  1023,  1023,  1023,  1022,  1022,  1022,  1022,  1022,
 1022,  1021,  1021,  1021,  1021,  1020,  1020,  1020,  1020,  1019,
 1019,  1019,  1019,  1018,  1018,  1018,  1017,  1017,  1017,  1016,
 1016,  1015,  1015,  1015,  1014,  1014,  1013,  1013,  1012,  1012,
 1011,  1011,  1010,  1010,  1009,  1009,  1008,  1008,  1007,  1007,
 1006,  1006,  1005,  1004,  1004,  1003,  1003,  1002,  1001,  1001,
 1000,   999,   999,   998,   997,   997,   996,   995,   994,   994,
  993,   992,   991,   990,   990,   989,   988,   987,   986,   986,
  985,   984,   983,   982,   981,   980,   979,   978,   978,   977,
  976,   975,   974,   973,   972,   971,   970,   969,   968,   967,
  966,   965,   964,   963,   962,   960,   959,   958,   957,   956,
  955,   954,   953,   951,   950,   949,   948,   947,   946,   944,
  943,   942,   941,   939,   938,   937,   936,   934,   933,   932,
  930,   929,   928,   927,   925,   924,   922,   921,   920,   918,
  917,   916,   914,   913,   911,   910,   908,   907,   906,   904,
  903,   901,   900,   898,   897,   895,   894,   892,   890,   889,
  887,   886,   884,   883,   881,   879,   878,   876,   875,   873,
  871,   870,   868,   866,   865,   863,   861,   860,   858,   856,
  854,   853,   851,   849,   847,   846,   844,   842,   840,   839,
  837,   835,   833,   831,   829,   828,   826,   824,   822,   820,
  818,   816,   814,   813,   811,   809,   807,   805,   803,   801,
  799,   797,   795,   793,   791,   789,   787,   785,   783,   781,
  779,   777,   775,   773,   771,   769,   767,   765,   762,   760,
  758,   756,   754,   752,   750,   748,   745,   743,   741,   739,
  737,   735,   732,   730,   728,   726,   724,   721,   719,   717,
  715,   712,   710,   708,   706,   703,   701,   699,   696,   694,
  692,   690,   687,   685,   683,   680,   678,   675,   673,   671,
  668,   666,   664,   661,   659,   656,   654,   652,   649,   647,
  644,   642,   639,   637,   634,   632,   629,   627,   625,   622,
  620,   617,   615,   612,   609,   607,   604,   602,   599,   597,
  594,   592,   589,   587,   584,   581,   579,   576,   574,   571,
  568,   566,   563,   561,   558,   555,   553,   550,   547,   545,
  542,   539,   537,   534,   531,   529,   526,   523,   521,   518,
  515,   512,   510,   507,   504,   501,   499,   496,   493,   491,
  488,   485,   482,   479,   477,   474,   471,   468,   466,   463,
  460,   457,   454,   451,   449,   446,   443,   440,   437,   434,
  432,   429,   426,   423,   420,   417,   414,   412,   409,   406,
  403,   400,   397,   394,   391,   388,   386,   383,   380,   377,
  374,   371,   368,   365,   362,   359,   356,   353,   350,   347,
  344,   342,   339,   336,   333,   330,   327,   324,   321,   318,
  315,   312,   309,   306,   303,   300,   297,   294,   291,   288,
  285,   282,   279,   276,   273,   270,   267,   264,   260,   257,
  254,   251,   248,   245,   242,   239,   236,   233,   230,   227,
  224,   221,   218,   215,   212,   209,   205,   202,   199,   196,
  193,   190,   187,   184,   181,   178,   175,   171,   168,   165,
  162,   159,   156,   153,   150,   147,   144,   140,   137,   134,
  131,   128,   125,   122,   119,   115,   112,   109,   106,   103,
  100,    97,    94,    90,    87,    84,    81,    78,    75,    72,
   69,    65,    62,    59,    56,    53,    50,    47,    43,    40,
   37,    34,    31,    28,    25,    21,    18,    15,    12,     9,
    6,     3,     0,    -3,    -6,    -9,   -12,   -15,   -18,   -21,
  -25,   -28,   -31,   -34,   -37,   -40,   -43,   -47,   -50,   -53,
  -56,   -59,   -62,   -65,   -69,   -72,   -75,   -78,   -81,   -84,
  -87,   -90,   -94,   -97,  -100,  -103,  -106,  -109,  -112,  -115,
 -119,  -122,  -125,  -128,  -131,  -134,  -137,  -140,  -144,  -147,
 -150,  -153,  -156,  -159,  -162,  -165,  -168,  -171,  -175,  -178,
 -181,  -184,  -187,  -190,  -193,  -196,  -199,  -202,  -205,  -209,
 -212,  -215,  -218,  -221,  -224,  -227,  -230,  -233,  -236,  -239,
 -242,  -245,  -248,  -251,  -254,  -257,  -260,  -264,  -267,  -270,
 -273,  -276,  -279,  -282,  -285,  -288,  -291,  -294,  -297,  -300,
 -303,  -306,  -309,  -312,  -315,  -318,  -321,  -324,  -327,  -330,
 -333,  -336,  -339,  -342,  -344,  -347,  -350,  -353,  -356,  -359,
 -362,  -365,  -368,  -371,  -374,  -377,  -380,  -383,  -386,  -388,
 -391,  -394,  -397,  -400,  -403,  -406,  -409,  -412,  -414,  -417,
 -420,  -423,  -426,  -429,  -432,  -434,  -437,  -440,  -443,  -446,
 -449,  -451,  -454,  -457,  -460,  -463,  -466,  -468,  -471,  -474,
 -477,  -479,  -482,  -485,  -488,  -491,  -493,  -496,  -499,  -501,
 -504,  -507,  -510,  -512,  -515,  -518,  -521,  -523,  -526,  -529,
 -531,  -534,  -537,  -539,  -542,  -545,  -547,  -550,  -553,  -555,
 -558,  -561,  -563,  -566,  -568,  -571,  -574,  -576,  -579,  -581,
 -584,  -587,  -589,  -592,  -594,  -597,  -599,  -602,  -604,  -607,
 -609,  -612,  -615,  -617,  -620,  -622,  -625,  -627,  -629,  -632,
 -634,  -637,  -639,  -642,  -644,  -647,  -649,  -652,  -654,  -656,
 -659,  -661,  -664,  -666,  -668,  -671,  -673,  -675,  -678,  -680,
 -683,  -685,  -687,  -689,  -692,  -694,  -696,  -699,  -701,  -703,
 -706,  -708,  -710,  -712,  -715,  -717,  -719,  -721,  -724,  -726,
 -728,  -730,  -732,  -735,  -737,  -739,  -741,  -743,  -745,  -748,
 -750,  -752,  -754,  -756,  -758,  -760,  -762,  -765,  -767,  -769,
 -771,  -773,  -775,  -777,  -779,  -781,  -783,  -785,  -787,  -789,
 -791,  -793,  -795,  -797,  -799,  -801,  -803,  -805,  -807,  -809,
 -811,  -813,  -814,  -816,  -818,  -820,  -822,  -824,  -826,  -828,
 -829,  -831,  -833,  -835,  -837,  -839,  -840,  -842,  -844,  -846,
 -847,  -849,  -851,  -853,  -854,  -856,  -858,  -860,  -861,  -863,
 -865,  -866,  -868,  -870,  -871,  -873,  -875,  -876,  -878,  -879,
 -881,  -883,  -884,  -886,  -887,  -889,  -890,  -892,  -894,  -895,
 -897,  -898,  -900,  -901,  -903,  -904,  -906,  -907,  -908,  -910,
 -911,  -913,  -914,  -916,  -917,  -918,  -920,  -921,  -922,  -924,
 -925,  -927,  -928,  -929,  -930,  -932,  -933,  -934,  -936,  -937,
 -938,  -939,  -941,  -942,  -943,  -944,  -946,  -947,  -948,  -949,
 -950,  -951,  -953,  -954,  -955,  -956,  -957,  -958,  -959,  -960,
 -962,  -963,  -964,  -965,  -966,  -967,  -968,  -969,  -970,  -971,
 -972,  -973,  -974,  -975,  -976,  -977,  -978,  -978,  -979,  -980,
 -981,  -982,  -983,  -984,  -985,  -986,  -986,  -987,  -988,  -989,
 -990,  -990,  -991,  -992,  -993,  -994,  -994,  -995,  -996,  -997,
 -997,  -998,  -999,  -999, -1000, -1001, -1001, -1002, -1003, -1003,
-1004, -1004, -1005, -1006, -1006, -1007, -1007, -1008, -1008, -1009,
-1009, -1010, -1010, -1011, -1011, -1012, -1012, -1013, -1013, -1014,
-1014, -1015, -1015, -1015, -1016, -1016, -1017, -1017, -1017, -1018,
-1018, -1018, -1019, -1019, -1019, -1019, -1020, -1020, -1020, -1020,
-1021, -1021, -1021, -1021, -1022, -1022, -1022, -1022, -1022, -1022,
-1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023,
-1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023,
-1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023, -1023, -1022,
-1022, -1022, -1022, -1022, -1022, -1021, -1021, -1021, -1021, -1020,
-1020, -1020, -1020, -1019, -1019, -1019, -1019, -1018, -1018, -1018,
-1017, -1017, -1017, -1016, -1016, -1015, -1015, -1015, -1014, -1014,
-1013, -1013, -1012, -1012, -1011, -1011, -1010, -1010, -1009, -1009,
-1008, -1008, -1007, -1007, -1006, -1006, -1005, -1004, -1004, -1003,
-1003, -1002, -1001, -1001, -1000,  -999,  -999,  -998,  -997,  -997,
 -996,  -995,  -994,  -994,  -993,  -992,  -991,  -990,  -990,  -989,
 -988,  -987,  -986,  -986,  -985,  -984,  -983,  -982,  -981,  -980,
 -979,  -978,  -978,  -977,  -976,  -975,  -974,  -973,  -972,  -971,
 -970,  -969,  -968,  -967,  -966,  -965,  -964,  -963,  -962,  -960,
 -959,  -958,  -957,  -956,  -955,  -954,  -953,  -951,  -950,  -949,
 -948,  -947,  -946,  -944,  -943,  -942,  -941,  -939,  -938,  -937,
 -936,  -934,  -933,  -932,  -930,  -929,  -928,  -927,  -925,  -924,
 -922,  -921,  -920,  -918,  -917,  -916,  -914,  -913,  -911,  -910,
 -908,  -907,  -906,  -904,  -903,  -901,  -900,  -898,  -897,  -895,
 -894,  -892,  -890,  -889,  -887,  -886,  -884,  -883,  -881,  -879,
 -878,  -876,  -875,  -873,  -871,  -870,  -868,  -866,  -865,  -863,
 -861,  -860,  -858,  -856,  -854,  -853,  -851,  -849,  -847,  -846,
 -844,  -842,  -840,  -839,  -837,  -835,  -833,  -831,  -829,  -828,
 -826,  -824,  -822,  -820,  -818,  -816,  -814,  -813,  -811,  -809,
 -807,  -805,  -803,  -801,  -799,  -797,  -795,  -793,  -791,  -789,
 -787,  -785,  -783,  -781,  -779,  -777,  -775,  -773,  -771,  -769,
 -767,  -765,  -762,  -760,  -758,  -756,  -754,  -752,  -750,  -748,
 -745,  -743,  -741,  -739,  -737,  -735,  -732,  -730,  -728,  -726,
 -724,  -721,  -719,  -717,  -715,  -712,  -710,  -708,  -706,  -703,
 -701,  -699,  -696,  -694,  -692,  -690,  -687,  -685,  -683,  -680,
 -678,  -675,  -673,  -671,  -668,  -666,  -664,  -661,  -659,  -656,
 -654,  -652,  -649,  -647,  -644,  -642,  -639,  -637,  -634,  -632,
 -629,  -627,  -625,  -622,  -620,  -617,  -615,  -612,  -609,  -607,
 -604,  -602,  -599,  -597,  -594,  -592,  -589,  -587,  -584,  -581,
 -579,  -576,  -574,  -571,  -568,  -566,  -563,  -561,  -558,  -555,
 -553,  -550,  -547,  -545,  -542,  -539,  -537,  -534,  -531,  -529,
 -526,  -523,  -521,  -518,  -515,  -512,  -510,  -507,  -504,  -501,
 -499,  -496,  -493,  -491,  -488,  -485,  -482,  -479,  -477,  -474,
 -471,  -468,  -466,  -463,  -460,  -457,  -454,  -451,  -449,  -446,
 -443,  -440,  -437,  -434,  -432,  -429,  -426,  -423,  -420,  -417,
 -414,  -412,  -409,  -406,  -403,  -400,  -397,  -394,  -391,  -388,
 -386,  -383,  -380,  -377,  -374,  -371,  -368,  -365,  -362,  -359,
 -356,  -353,  -350,  -347,  -344,  -342,  -339,  -336,  -333,  -330,
 -327,  -324,  -321,  -318,  -315,  -312,  -309,  -306,  -303,  -300,
 -297,  -294,  -291,  -288,  -285,  -282,  -279,  -276,  -273,  -270,
 -267,  -264,  -260,  -257,  -254,  -251,  -248,  -245,  -242,  -239,
 -236,  -233,  -230,  -227,  -224,  -221,  -218,  -215,  -212,  -209,
 -205,  -202,  -199,  -196,  -193,  -190,  -187,  -184,  -181,  -178,
 -175,  -171,  -168,  -165,  -162,  -159,  -156,  -153,  -150,  -147,
 -144,  -140,  -137,  -134,  -131,  -128,  -125,  -122,  -119,  -115,
 -112,  -109,  -106,  -103,  -100,   -97,   -94,   -90,   -87,   -84,
  -81,   -78,   -75,   -72,   -69,   -65,   -62,   -59,   -56,   -53,
  -50,   -47,   -43,   -40,   -37,   -34,   -31,   -28,   -25,   -21,
  -18,   -15,   -12,    -9,    -6,    -3,     0,     3,     6,     9,
   12,    15,    18,    21,    25,    28,    31,    34,    37,    40,
   43,    47,    50,    53,    56,    59,    62,    65,    69,    72,
   75,    78,    81,    84,    87,    90,    94,    97,   100,   103,
  106,   109,   112,   115,   119,   122,   125,   128,   131,   134,
  137,   140,   144,   147,   150,   153,   156,   159,   162,   165,
  168,   171,   175,   178,   181,   184,   187,   190,   193,   196,
  199,   202,   205,   209,   212,   215,   218,   221,   224,   227,
  230,   233,   236,   239,   242,   245,   248,   251,   254,   257,
  260,   264,   267,   270,   273,   276,   279,   282,   285,   288,
  291,   294,   297,   300,   303,   306,   309,   312,   315,   318,
  321,   324,   327,   330,   333,   336,   339,   342,   344,   347,
  350,   353,   356,   359,   362,   365,   368,   371,   374,   377,
  380,   383,   386,   388,   391,   394,   397,   400,   403,   406,
  409,   412,   414,   417,   420,   423,   426,   429,   432,   434,
  437,   440,   443,   446,   449,   451,   454,   457,   460,   463,
  466,   468,   471,   474,   477,   479,   482,   485,   488,   490,
  493,   496,   499,   501,   504,   507,   510,   512,   515,   518,
  521,   523,   526,   529,   531,   534,   537,   539,   542,   545,
  547,   550,   553,   555,   558,   561,   563,   566,   568,   571,
  574,   576,   579,   581,   584,   587,   589,   592,   594,   597,
  599,   602,   604,   607,   609,   612,   615,   617,   620,   622,
  625,   627,   629,   632,   634,   637,   639,   642,   644,   647,
  649,   652,   654,   656,   659,   661,   664,   666,   668,   671,
  673,   675,   678,   680,   683,   685,   687,   689,   692,   694,
  696,   699,   701,   703,   706,   708,   710,   712,   715,   717,
  719,   721,   724,   726,   728,   730,   732,   735,   737,   739,
  741,   743,   745,   748,   750,   752,   754,   756,   758,   760,
  762,   765,   767,   769,   771,   773,   775,   777,   779,   781,
  783,   785,   787,   789,   791,   793,   795,   797,   799,   801,
  803,   805,   807,   809,   811,   813,   814,   816,   818,   820,
  822,   824,   826,   828,   829,   831,   833,   835,   837,   839,
  840,   842,   844,   846,   847,   849,   851,   853,   854,   856,
  858,   860,   861,   863,   865,   866,   868,   870,   871,   873,
  875,   876,   878,   879,   881,   883,   884,   886,   887,   889,
  890,   892,   894,   895,   897,   898,   900,   901,   903,   904,
  906,   907,   908,   910,   911,   913,   914,   916,   917,   918,
  920,   921,   922,   924,   925,   927,   928,   929,   930,   932,
  933,   934,   936,   937,   938,   939,   941,   942,   943,   944,
  946,   947,   948,   949,   950,   951,   953,   954,   955,   956,
  957,   958,   959,   960,   962,   963,   964,   965,   966,   967,
  968,   969,   970,   971,   972,   973,   974,   975,   976,   977,
  978,   978,   979,   980,   981,   982,   983,   984,   985,   986,
  986,   987,   988,   989,   990,   990,   991,   992,   993,   994,
  994,   995,   996,   997,   997,   998,   999,   999,  1000,  1001,
 1001,  1002,  1003,  1003,  1004,  1004,  1005,  1006,  1006,  1007,
 1007,  1008,  1008,  1009,  1009,  1010,  1010,  1011,  1011,  1012,
 1012,  1013,  1013,  1014,  1014,  1015,  1015,  1015,  1016,  1016,
 1017,  1017,  1017,  1018,  1018,  1018,  1019,  1019,  1019,  1019,
 1020,  1020,  1020,  1020,  1021,  1021,  1021,  1021,  1022,  1022,
 1022,  1022,  1022,  1022,  1023,  1023,  1023,  1023,  1023,  1023,
 1023,  1023,  1023,  1023,  1023,  1023,  1023,  1023
};

const Sfxp __TAN_TABLE[TABLE_SIZE] = {
         0,          3,          6,          9,         12,
        15,         18,         21,         25,         28,
        31,         34,         37,         40,         44,
        47,         50,         53,         56,         59,
        62,         66,         69,         72,         75,
        78,         81,         85,         88,         91,
        94,         97,        100,        104,        107,
       110,        113,        116,        119,        123,
       126,        129,        132,        135,        139,
       142,        145,        148,        151,        155,
       158,        161,        164,        167,        171,
       174,        177,        180,        184,        187,
       190,        193,        197,        200,        203,
       206,        210,        213,        216,        220,
       223,        226,        229,        233,        236,
       239,        243,        246,        249,        253,
       256,        259,        263,        266,        269,
       273,        276,        279,        283,        286,
       290,        293,        296,        300,        303,
       307,        310,        314,        317,        320,
       324,        327,        331,        334,        338,
       341,        345,        348,        352,        355,
       359,        362,        366,        369,        373,
       377,        380,        384,        387,        391,
       395,        398,        402,        405,        409,
       413,        416,        420,        424,        427,
       431,        435,        438,        442,        446,
       450,        453,        457,        461,        465,
       469,        472,        476,        480,        484,
       488,        492,        495,        499,        503,
       507,        511,        515,        519,        523,
       527,        531,        535,        539,        543,
       547,        551,        555,        559,        563,
       567,        571,        575,        580,        584,
       588,        592,        596,        601,        605,
       609,        613,        618,        622,        626,
       630,        635,        639,        644,        648,
       652,        657,        661,        666,        670,
       675,        679,        684,        688,        693,
       697,        702,        707,        711,        716,
       721,        725,        730,        735,        740,
       744,        749,        754,        759,        764,
       769,        774,        779,        784,        789,
       794,        799,        804,        809,        814,
       819,        824,        829,        835,        840,
       845,        850,        856,        861,        866,
       872,        877,        883,        888,        894,
       899,        905,        911,        916,        922,
       928,        933,        939,        945,        951,
       957,        963,        968,        974,        980,
       986,        993,        999,       1005,       1011,
      1017,       1023,       1030,       1036,       1043,
      1049,       1055,       1062,       1068,       1075,
      1082,       1088,       1095,       1102,       1109,
      1115,       1122,       1129,       1136,       1143,
      1150,       1158,       1165,       1172,       1179,
      1187,       1194,       1201,       1209,       1216,
      1224,       1232,       1239,       1247,       1255,
      1263,       1271,       1279,       1287,       1295,
      1303,       1312,       1320,       1328,       1337,
      1345,       1354,       1363,       1371,       1380,
      1389,       1398,       1407,       1416,       1425,
      1435,       1444,       1453,       1463,       1473,
      1482,       1492,       1502,       1512,       1522,
      1532,       1542,       1553,       1563,       1573,
      1584,       1595,       1606,       1617,       1628,
      1639,       1650,       1661,       1673,       1684,
      1696,       1708,       1720,       1732,       1744,
      1756,       1769,       1782,       1794,       1807,
      1820,       1833,       1847,       1860,       1874,
      1887,       1901,       1915,       1929,       1944,
      1958,       1973,       1988,       2003,       2018,
      2034,       2049,       2065,       2081,       2098,
      2114,       2131,       2147,       2165,       2182,
      2199,       2217,       2235,       2253,       2272,
      2291,       2310,       2329,       2348,       2368,
      2388,       2409,       2429,       2450,       2472,
      2493,       2515,       2537,       2560,       2583,
      2606,       2630,       2654,       2679,       2703,
      2729,       2754,       2780,       2807,       2834,
      2861,       2889,       2918,       2947,       2976,
      3006,       3036,       3068,       3099,       3131,
      3164,       3198,       3232,       3267,       3302,
      3338,       3375,       3413,       3451,       3491,
      3531,       3572,       3613,       3656,       3700,
      3744,       3790,       3837,       3885,       3934,
      3984,       4035,       4088,       4141,       4197,
      4253,       4311,       4371,       4432,       4495,
      4560,       4626,       4694,       4764,       4836,
      4911,       4987,       5066,       5147,       5231,
      5318,       5407,       5499,       5595,       5693,
      5795,       5901,       6010,       6124,       6241,
      6364,       6491,       6622,       6760,       6903,
      7052,       7207,       7369,       7539,       7716,
      7902,       8097,       8302,       8517,       8743,
      8981,       9233,       9499,       9781,      10079,
     10396,      10734,      11094,      11478,      11890,
     12333,      12809,      13324,      13881,      14487,
     15148,      15871,      16667,      17546,      18523,
     19615,      20843,      22235,      23825,      25660,
     27800,      30330,      33365,      37074,      41711,
     47671,      55618,      66743,      83429,     111238,
    166849,     333632,  807831678,    -333908,    -166917,
   -111269,     -83447,     -66754,     -55626,     -47677,
    -41715,     -37078,     -33368,     -30332,     -27802,
    -25661,     -23826,     -22236,     -20844,     -19616,
    -18524,     -17547,     -16668,     -15872,     -15148,
    -14488,     -13882,     -13324,     -12810,     -12333,
    -11891,     -11479,     -11094,     -10734,     -10396,
    -10079,      -9781,      -9499,      -9233,      -8982,
     -8743,      -8517,      -8302,      -8097,      -7903,
     -7717,      -7539,      -7370,      -7207,      -7052,
     -6903,      -6760,      -6623,      -6491,      -6364,
     -6242,      -6124,      -6010,      -5901,      -5795,
     -5693,      -5595,      -5499,      -5407,      -5318,
     -5231,      -5148,      -5066,      -4987,      -4911,
     -4837,      -4764,      -4694,      -4626,      -4560,
     -4495,      -4432,      -4371,      -4311,      -4253,
     -4197,      -4141,      -4088,      -4035,      -3984,
     -3934,      -3885,      -3837,      -3790,      -3744,
     -3700,      -3656,      -3613,      -3572,      -3531,
     -3491,      -3451,      -3413,      -3375,      -3338,
     -3302,      -3267,      -3232,      -3198,      -3164,
     -3131,      -3099,      -3068,      -3037,      -3006,
     -2976,      -2947,      -2918,      -2889,      -2861,
     -2834,      -2807,      -2780,      -2754,      -2729,
     -2703,      -2679,      -2654,      -2630,      -2606,
     -2583,      -2560,      -2537,      -2515,      -2493,
     -2472,      -2450,      -2429,      -2409,      -2388,
     -2368,      -2348,      -2329,      -2310,      -2291,
     -2272,      -2253,      -2235,      -2217,      -2199,
     -2182,      -2165,      -2148,      -2131,      -2114,
     -2098,      -2081,      -2065,      -2049,      -2034,
     -2018,      -2003,      -1988,      -1973,      -1958,
     -1944,      -1929,      -1915,      -1901,      -1887,
     -1874,      -1860,      -1847,      -1833,      -1820,
     -1807,      -1794,      -1782,      -1769,      -1756,
     -1744,      -1732,      -1720,      -1708,      -1696,
     -1684,      -1673,      -1661,      -1650,      -1639,
     -1628,      -1617,      -1606,      -1595,      -1584,
     -1574,      -1563,      -1553,      -1542,      -1532,
     -1522,      -1512,      -1502,      -1492,      -1482,
     -1473,      -1463,      -1453,      -1444,      -1435,
     -1425,      -1416,      -1407,      -1398,      -1389,
     -1380,      -1371,      -1363,      -1354,      -1345,
     -1337,      -1328,      -1320,      -1312,      -1303,
     -1295,      -1287,      -1279,      -1271,      -1263,
     -1255,      -1247,      -1239,      -1232,      -1224,
     -1216,      -1209,      -1201,      -1194,      -1187,
     -1179,      -1172,      -1165,      -1158,      -1150,
     -1143,      -1136,      -1129,      -1122,      -1115,
     -1109,      -1102,      -1095,      -1088,      -1082,
     -1075,      -1068,      -1062,      -1055,      -1049,
     -1043,      -1036,      -1030,      -1024,      -1017,
     -1011,      -1005,       -999,       -993,       -986,
      -980,       -974,       -968,       -963,       -957,
      -951,       -945,       -939,       -933,       -928,
      -922,       -916,       -911,       -905,       -899,
      -894,       -888,       -883,       -877,       -872,
      -867,       -861,       -856,       -850,       -845,
      -840,       -835,       -829,       -824,       -819,
      -814,       -809,       -804,       -799,       -794,
      -789,       -784,       -779,       -774,       -769,
      -764,       -759,       -754,       -749,       -744,
      -740,       -735,       -730,       -725,       -721,
      -716,       -711,       -707,       -702,       -697,
      -693,       -688,       -684,       -679,       -675,
      -670,       -666,       -661,       -657,       -652,
      -648,       -644,       -639,       -635,       -630,
      -626,       -622,       -618,       -613,       -609,
      -605,       -601,       -596,       -592,       -588,
      -584,       -580,       -575,       -571,       -567,
      -563,       -559,       -555,       -551,       -547,
      -543,       -539,       -535,       -531,       -527,
      -523,       -519,       -515,       -511,       -507,
      -503,       -499,       -495,       -492,       -488,
      -484,       -480,       -476,       -472,       -469,
      -465,       -461,       -457,       -453,       -450,
      -446,       -442,       -438,       -435,       -431,
      -427,       -424,       -420,       -416,       -413,
      -409,       -405,       -402,       -398,       -395,
      -391,       -387,       -384,       -380,       -377,
      -373,       -369,       -366,       -362,       -359,
      -355,       -352,       -348,       -345,       -341,
      -338,       -334,       -331,       -327,       -324,
      -320,       -317,       -314,       -310,       -307,
      -303,       -300,       -296,       -293,       -290,
      -286,       -283,       -280,       -276,       -273,
      -269,       -266,       -263,       -259,       -256,
      -253,       -249,       -246,       -243,       -239,
      -236,       -233,       -229,       -226,       -223,
      -220,       -216,       -213,       -210,       -206,
      -203,       -200,       -197,       -193,       -190,
      -187,       -184,       -180,       -177,       -174,
      -171,       -167,       -164,       -161,       -158,
      -155,       -151,       -148,       -145,       -142,
      -139,       -135,       -132,       -129,       -126,
      -123,       -119,       -116,       -113,       -110,
      -107,       -104,       -100,        -97,        -94,
       -91,        -88,        -85,        -81,        -78,
       -75,        -72,        -69,        -66,        -62,
       -59,        -56,        -53,        -50,        -47,
       -44,        -40,        -37,        -34,        -31,
       -28,        -25,        -21,        -18,        -15,
       -12,         -9,         -6,         -3,          0,
         3,          6,          9,         12,         15,
        18,         21,         25,         28,         31,
        34,         37,         40,         44,         47,
        50,         53,         56,         59,         62,
        66,         69,         72,         75,         78,
        81,         85,         88,         91,         94,
        97,        100,        104,        107,        110,
       113,        116,        119,        123,        126,
       129,        132,        135,        139,        142,
       145,        148,        151,        155,        158,
       161,        164,        167,        171,        174,
       177,        180,        184,        187,        190,
       193,        197,        200,        203,        206,
       210,        213,        216,        220,        223,
       226,        229,        233,        236,        239,
       243,        246,        249,        253,        256,
       259,        263,        266,        269,        273,
       276,        279,        283,        286,        290,
       293,        296,        300,        303,        307,
       310,        314,        317,        320,        324,
       327,        331,        334,        338,        341,
       345,        348,        352,        355,        359,
       362,        366,        369,        373,        377,
       380,        384,        387,        391,        394,
       398,        402,        405,        409,        413,
       416,        420,        424,        427,        431,
       435,        438,        442,        446,        450,
       453,        457,        461,        465,        469,
       472,        476,        480,        484,        488,
       492,        495,        499,        503,        507,
       511,        515,        519,        523,        527,
       531,        535,        539,        543,        547,
       551,        555,        559,        563,        567,
       571,        575,        580,        584,        588,
       592,        596,        601,        605,        609,
       613,        618,        622,        626,        630,
       635,        639,        644,        648,        652,
       657,        661,        666,        670,        675,
       679,        684,        688,        693,        697,
       702,        707,        711,        716,        721,
       725,        730,        735,        740,        744,
       749,        754,        759,        764,        769,
       774,        779,        784,        789,        794,
       799,        804,        809,        814,        819,
       824,        829,        835,        840,        845,
       850,        856,        861,        866,        872,
       877,        883,        888,        894,        899,
       905,        911,        916,        922,        928,
       933,        939,        945,        951,        957,
       963,        968,        974,        980,        986,
       993,        999,       1005,       1011,       1017,
      1023,       1030,       1036,       1043,       1049,
      1055,       1062,       1068,       1075,       1082,
      1088,       1095,       1102,       1109,       1115,
      1122,       1129,       1136,       1143,       1150,
      1158,       1165,       1172,       1179,       1187,
      1194,       1201,       1209,       1216,       1224,
      1232,       1239,       1247,       1255,       1263,
      1271,       1279,       1287,       1295,       1303,
      1312,       1320,       1328,       1337,       1345,
      1354,       1363,       1371,       1380,       1389,
      1398,       1407,       1416,       1425,       1435,
      1444,       1453,       1463,       1473,       1482,
      1492,       1502,       1512,       1522,       1532,
      1542,       1553,       1563,       1573,       1584,
      1595,       1606,       1617,       1628,       1639,
      1650,       1661,       1673,       1684,       1696,
      1708,       1720,       1732,       1744,       1756,
      1769,       1782,       1794,       1807,       1820,
      1833,       1847,       1860,       1874,       1887,
      1901,       1915,       1929,       1944,       1958,
      1973,       1988,       2003,       2018,       2034,
      2049,       2065,       2081,       2098,       2114,
      2131,       2147,       2165,       2182,       2199,
      2217,       2235,       2253,       2272,       2291,
      2310,       2329,       2348,       2368,       2388,
      2409,       2429,       2450,       2472,       2493,
      2515,       2537,       2560,       2583,       2606,
      2630,       2654,       2679,       2703,       2729,
      2754,       2780,       2807,       2834,       2861,
      2889,       2918,       2947,       2976,       3006,
      3036,       3068,       3099,       3131,       3164,
      3198,       3232,       3267,       3302,       3338,
      3375,       3413,       3451,       3490,       3531,
      3571,       3613,       3656,       3700,       3744,
      3790,       3837,       3885,       3933,       3984,
      4035,       4087,       4141,       4197,       4253,
      4311,       4371,       4432,       4495,       4560,
      4626,       4694,       4764,       4836,       4911,
      4987,       5066,       5147,       5231,       5318,
      5407,       5499,       5595,       5693,       5795,
      5901,       6010,       6124,       6241,       6363,
      6490,       6622,       6760,       6903,       7052,
      7207,       7369,       7539,       7716,       7902,
      8097,       8302,       8517,       8743,       8981,
      9233,       9499,       9780,      10079,      10396,
     10733,      11093,      11478,      11890,      12333,
     12809,      13324,      13881,      14487,      15147,
     15870,      16666,      17545,      18522,      19614,
     20842,      22233,      23824,      25658,      27798,
     30327,      33362,      37071,      41706,      47665,
     55610,      66732,      83412,     111207,     166780,
    333356,  269277226,    -334184,    -166986,    -111299,
    -83464,     -66765,     -55633,     -47682,     -41719,
    -37081,     -33370,     -30334,     -27804,     -25663,
    -23828,     -22237,     -20845,     -19617,     -18525,
    -17548,     -16668,     -15872,     -15149,     -14488,
    -13882,     -13325,     -12810,     -12334,     -11891,
    -11479,     -11094,     -10734,     -10397,     -10080,
     -9781,      -9500,      -9234,      -8982,      -8743,
     -8517,      -8302,      -8098,      -7903,      -7717,
     -7539,      -7370,      -7207,      -7052,      -6903,
     -6760,      -6623,      -6491,      -6364,      -6242,
     -6124,      -6011,      -5901,      -5795,      -5693,
     -5595,      -5500,      -5407,      -5318,      -5231,
     -5148,      -5066,      -4987,      -4911,      -4837,
     -4764,      -4694,      -4626,      -4560,      -4495,
     -4432,      -4371,      -4311,      -4253,      -4197,
     -4141,      -4088,      -4035,      -3984,      -3934,
     -3885,      -3837,      -3790,      -3744,      -3700,
     -3656,      -3613,      -3572,      -3531,      -3491,
     -3451,      -3413,      -3375,      -3338,      -3302,
     -3267,      -3232,      -3198,      -3164,      -3131,
     -3099,      -3068,      -3037,      -3006,      -2976,
     -2947,      -2918,      -2889,      -2861,      -2834,
     -2807,      -2780,      -2754,      -2729,      -2703,
     -2679,      -2654,      -2630,      -2606,      -2583,
     -2560,      -2538,      -2515,      -2493,      -2472,
     -2450,      -2429,      -2409,      -2388,      -2368,
     -2348,      -2329,      -2310,      -2291,      -2272,
     -2253,      -2235,      -2217,      -2199,      -2182,
     -2165,      -2148,      -2131,      -2114,      -2098,
     -2081,      -2065,      -2049,      -2034,      -2018,
     -2003,      -1988,      -1973,      -1958,      -1944,
     -1930,      -1915,      -1901,      -1887,      -1874,
     -1860,      -1847,      -1833,      -1820,      -1807,
     -1794,      -1782,      -1769,      -1756,      -1744,
     -1732,      -1720,      -1708,      -1696,      -1684,
     -1673,      -1661,      -1650,      -1639,      -1628,
     -1617,      -1606,      -1595,      -1584,      -1574,
     -1563,      -1553,      -1542,      -1532,      -1522,
     -1512,      -1502,      -1492,      -1482,      -1473,
     -1463,      -1453,      -1444,      -1435,      -1425,
     -1416,      -1407,      -1398,      -1389,      -1380,
     -1371,      -1363,      -1354,      -1345,      -1337,
     -1328,      -1320,      -1312,      -1303,      -1295,
     -1287,      -1279,      -1271,      -1263,      -1255,
     -1247,      -1239,      -1232,      -1224,      -1216,
     -1209,      -1201,      -1194,      -1187,      -1179,
     -1172,      -1165,      -1158,      -1150,      -1143,
     -1136,      -1129,      -1122,      -1115,      -1109,
     -1102,      -1095,      -1088,      -1082,      -1075,
     -1068,      -1062,      -1055,      -1049,      -1043,
     -1036,      -1030,      -1024,      -1017,      -1011,
     -1005,       -999,       -993,       -986,       -980,
      -974,       -968,       -963,       -957,       -951,
      -945,       -939,       -933,       -928,       -922,
      -916,       -911,       -905,       -899,       -894,
      -888,       -883,       -877,       -872,       -867,
      -861,       -856,       -850,       -845,       -840,
      -835,       -829,       -824,       -819,       -814,
      -809,       -804,       -799,       -794,       -789,
      -784,       -779,       -774,       -769,       -764,
      -759,       -754,       -749,       -744,       -740,
      -735,       -730,       -725,       -721,       -716,
      -711,       -707,       -702,       -697,       -693,
      -688,       -684,       -679,       -675,       -670,
      -666,       -661,       -657,       -652,       -648,
      -644,       -639,       -635,       -630,       -626,
      -622,       -618,       -613,       -609,       -605,
      -601,       -596,       -592,       -588,       -584,
      -580,       -575,       -571,       -567,       -563,
      -559,       -555,       -551,       -547,       -543,
      -539,       -535,       -531,       -527,       -523,
      -519,       -515,       -511,       -507,       -503,
      -499,       -495,       -492,       -488,       -484,
      -480,       -476,       -472,       -469,       -465,
      -461,       -457,       -453,       -450,       -446,
      -442,       -438,       -435,       -431,       -427,
      -424,       -420,       -416,       -413,       -409,
      -405,       -402,       -398,       -395,       -391,
      -387,       -384,       -380,       -377,       -373,
      -369,       -366,       -362,       -359,       -355,
      -352,       -348,       -345,       -341,       -338,
      -334,       -331,       -327,       -324,       -320,
      -317,       -314,       -310,       -307,       -303,
      -300,       -296,       -293,       -290,       -286,
      -283,       -280,       -276,       -273,       -269,
      -266,       -263,       -259,       -256,       -253,
      -249,       -246,       -243,       -239,       -236,
      -233,       -229,       -226,       -223,       -220,
      -216,       -213,       -210,       -206,       -203,
      -200,       -197,       -193,       -190,       -187,
      -184,       -180,       -177,       -174,       -171,
      -167,       -164,       -161,       -158,       -155,
      -151,       -148,       -145,       -142,       -139,
      -135,       -132,       -129,       -126,       -123,
      -119,       -116,       -113,       -110,       -107,
      -104,       -100,        -97,        -94,        -91,
       -88,        -85,        -81,        -78,        -75,
       -72,        -69,        -66,        -62,        -59,
       -56,        -53,        -50,        -47,        -44,
       -40,        -37,        -34,        -31,        -28,
       -25,        -21,        -18,        -15,        -12,
        -9,         -6,         -3
};
