/***************************************************************************
               sprite.h  -  Sprite types, and class definition
                         -------------------
    begin                : Wed Mar 5 2003
    copyright            : (C) 2003 by Tyler Montbriand
    email                : tsm@accesscomm.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef SPRITE_H
#define SPRITE_H
#include "RotateImage.h"
#include "Fxp.h"
#include "rect.h"
#include "bitmask.h"

/**
  *@author Tyler Montbriand
  */

#include <SDL/begin_code.h>

/*! \brief Type used for sprite status flag */
typedef Uint16  SpriteFlag;

/*! \brief Type used for collision layer */
typedef Uint8  ColType;

/*! \brief Converts a collision layer to a collision mask */
#define COL_LAYER(x)  (0xff&(1<<x))


/*! \brief No flags */
#define SPRITE_NONE       0x0000
/*! \brief Sprite is waiting for first draw or removal from list */
#define SPRITE_WAIT       0x0001
/*! \brief Sprite needs redraw */
#define SPRITE_MOVED      0x0002
/*! \brief Sprite needs removal from list */
#define SPRITE_EXPIRED    0x0004
/*! \brief Sprite can be freed from mem upon removal from list */  
#define SPRITE_DELETEABLE 0x0008
/*! \brief Sprite owns aka can free the contained SDL_surface */
#define SPRITE_OWNSURFACE 0x0010
/*! \brief Sprite owns aka can free the contained bitmasks */
#define SPRITE_OWNBITMASK 0x0020

/*! \brief The basic Sprite class itself  */
class DECLSPEC Sprite
{
  /*! \brief SpriteList has access to Sprite's private members    */
  friend class SpriteList;
  /*! \brief SpriteManager has access to Sprite's private members */
  friend class SpriteManager;
  
public:

  /*! \brief Default constructor */
  Sprite();  

  /*! \brief Calculates movement for the # of milliseconds */
  virtual void Update(Sint32 dtick);

  /*! \brief Called upon collision with another sprite */
  virtual void OnCollide(const Sprite *spr, ColType index);

  /*! \brief Returns the current frame number. */
  virtual int GetFrame();

  /*! \brief Returns a clone of the current sprite */
  virtual Sprite *Clone();

  /*! \brief Called for SDL_USEREVENTs  */
  virtual void Event(int code, void *data1=NULL, void *data2=NULL);

  inline SfxpCoord Pos()  const {return pos;  }
  inline SfxpCoord Vel()  const {return vel;  }

  /*! \brief Sets the frame rate. */
  inline void SetFrameRate(Sfxp frate)
  { framerate=frate;  }

  /*! \brief Marks a sprite as expired */
  inline void Remove()
  {  SetFlags(SPRITE_EXPIRED);  }

  /*! \brief Marks a sprite as moved */
  inline void Tag()
  {  SetFlags(SPRITE_MOVED);  }

  /*! \brief Returns true if the sprite can free it's surface */
  inline int OwnSurface() const
  {  return(flag&SPRITE_OWNSURFACE);  }

  /*! \brief Returns true if the sprite can free it's collision bitmasks  */
  inline int OwnBitmask() const
  {  return(flag&SPRITE_OWNBITMASK);  }

  /*! \brief Sets flags in sprite */
  inline SpriteFlag SetFlags(const SpriteFlag flags)
  {
    flag|=flags;
    return(flag);
  }
  /*! \brief Returns the sprite's drawing layer */  
  inline Sint16 DrawLayer() const
  { return(drawlayer);  }
  
  /*! \brief this is true when the sprite is deleteable */
  inline int Deleteable() const
  {  return(flag&SPRITE_DELETEABLE);  }

  /*! \brief This is true when the sprite is newly created, not having been drawn*/
  inline int New() const
  {  return((flag&SPRITE_WAIT)&&!(flag&SPRITE_EXPIRED));  }

  /*! \brief Returns whether movement flag's been set */
  inline int Moved() const
  {  return(flag&(SPRITE_MOVED|SPRITE_EXPIRED));  }

  /*! \brief Marks the sprite as removable from a list  */
  inline void Expire()
  {
    SetFlags(SPRITE_EXPIRED);
  }

  /*! \brief Returns true if the sprite is removable from a list  */
  inline int Expired() const
  {  return(flag&SPRITE_EXPIRED);  }

  /*! \brief Sets the sprite's velocity */
  inline void SetVel(SfxpCoord velin)
  {  vel=velin;  }
  
  /*! \brief Clears flags in sprite */
  inline const SpriteFlag ClearFlags(const SpriteFlag flags)
  {
    flag&=~flags;
    return(flag);
  }

  /*! \brief Calculate the source-rect for the current frame # */
  inline void CalcFrame()
  {
    int f=GetFrame();
    int xf=f%frames.x;
    int yf=(f/frames.x)%frames.y;

    srcrect.x=xf*frames.w;
    srcrect.y=yf*frames.h;
    srcrect.w=frames.w;
    srcrect.h=frames.h;
  }

  /*! \brief Completely reverses X direction  */
  inline void ReverseXVelocity()
  { vel.SetFxp(  -vel.Xfxp(), vel.Yfxp()  );  }
  
  /*! \brief Completely reverses Y direction  */
  inline void ReverseYVelocity()
  { vel.SetFxp(   vel.Xfxp(),-vel.Yfxp()  );  }
  

  /*! \brief Moves the sprite to a different position */
  inline void MoveTo(Sfxp x, Sfxp y, bool save_prev_pos=true)
  {
    if(save_prev_pos) prevpos=pos;
    
    pos.SetFxp(x,y);
  }

  /*! \brief Returns flags in sprite */
  inline const SpriteFlag GetFlags() const
  { return(flag);  }

  /*! \brief Returns last frame's position rect */
  Rect PrevRect() const
  {  return(Rect(prevpos,srcrect.w,srcrect.h)); }

  /*! \brief Returns current frame's position rect */
  Rect CurRect() const
  {  return(Rect(pos, srcrect.w,srcrect.h));    }

  /*! \brief Returns fixed-point x velocity */
  inline Sfxp VXfxp() const { return(vel.Xfxp()); }
  /*! \brief Returns fixed-point y velicity */  
  inline Sfxp VYfxp() const { return(vel.Yfxp()); }

  /*! \brief Returns fixed-point X position */
  inline Sfxp Xfxp()  const { return pos.Xfxp(); };
  /*! \brief Returns fixed-point Y position */
  inline Sfxp Yfxp()  const { return pos.Yfxp(); }

  /*! \brief Returns the current collision bitmask  */
  bitmask *GetBitmask();
  /*! \brief Returns the collision index mask       */
  inline Uint8 const GetColIndex()  { return(cindex); }
  /*! \brief Returns the collision layer mask       */
  inline ColType const GetColLayer(){ return(clayer); }

  /*! \brief Returns true if the 2 sprites collide, false otherwise */
  inline bool CanCollide(Sprite *spr)
  {    
    if(spr==NULL) return false;
    Rect r1=CurRect(),r2=spr->CurRect();
    
    if((clayer&spr->cindex)&&(cindex&spr->clayer))
    if(r1.Overlaps(r2))
    {
      bitmask *b1=GetBitmask(),*b2=spr->GetBitmask();
      
      if((b1!=NULL)&&(b2!=NULL))
      {
        int dx=r2.x-r1.x,dy=r2.y-r1.y;
        if(bitmask_overlap(b1,b2,dx,dy))/* Bitmask collision */
        {
          return(true);
        }
        else
        {
          return(false);
        }
      }
      else  /* Bounding box collision only */
      {
        return(true);
      }      
    }
    
    return false;
  }
  

  /*! \brief Draws the sprite on the screen               */
  void Draw(SDL_Surface *screen);
  /*! \brief Initializes the sprite's surface member      */
  void Init_Surface(SDL_Surface *surface=NULL, bool os_in=false);
  /*! \brief Initializes the sprite's drawing layer       */
  void Init_Layers(Sint16 layer=0);
  /*! \brief Initializes the sprite's linked-list members */
  void Init_List(Sprite *previn=NULL, Sprite *nextin=NULL);
  /*! \brief Initializes the sprite's status flags        */
  void Init_Flags(SpriteFlag flags=0);
  /*! \brief Initializes the sprite's animation frames    */
  void Init_Frames(int xframes=-1, int yframes=-1);
  /*! \brief Initializes the sprite's collision masks     */
  void Init_Collision(ColType c_index=0, ColType c_layer=0);
  /*! \brief Initializes the sprite's collision masks     */
  void Init_Bitmask(bool enable=false, bitmask **bmask_in=NULL);

  /*! \brief Destructor, frees any owned data */
  virtual ~Sprite();
  
protected:
  /*! \brief Collision bitmasks */
  bitmask **bmask;

  /*! \brief Collision layer mask, the layers it can collide with */
  ColType clayer;
  /*! \brief Collision layer index, the collision layer inhabited */
  ColType cindex;

  /*! \brief Current frame # in fixed point */
  Sfxp frame;
  /*! \brief Current framerate in fixed point */
  Sfxp framerate;
  /*! \brief x - x frames, y - y frames */    
  Rect   frames;
  
  /*! \brief Linked-list prev/next pointers */
  Sprite *prev,*next;
  
  /*! \brief Current position coordinates */
  SfxpCoord pos;
  /*! \brief Previous position coordinates */
  SfxpCoord prevpos;
  /*! \brief Velocity vector */
  SfxpCoord vel;

  /*! \brief Image surface */
  SDL_Surface *surface;
  /*! \brief Source rectangle */
  Rect srcrect;

  /*! \brief Higher #s get drawn first */
  Sint16 drawlayer;

  /*! \brief Status flags */
  SpriteFlag flag;
};


/*! \brief Loads an image and converts it to screen parameters  */
DECLSPEC SDL_Surface * SDLCALL LoadConvertBMP(const char *, Uint8 tr=255, Uint8 tg=0,
                            Uint8 tb=255                            );
/*! \brief Converts selected portion of image into collision bitmask */
DECLSPEC bitmask * SDLCALL MakeBitMask(SDL_Surface *surf, Rect rect);                            

#include <SDL/close_code.h>

#endif/*SPRITE_H*/


