/***************************************************************************
                          spriteloader.h  -  description
                             -------------------
    begin                : Thu Apr 10 2003
    copyright            : (C) 2003 by Tyler Montbriand
    email                : tsm@accesscomm.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef SPRITELOADER_H
#define SPRITELOADER_H

#include <malloc.h>
#include "RotateImage.h"
#include "sprite.h"

#include <SDL/begin_code.h>

/**
  *@author Tyler Montbriand
  */

#if defined(WIN32) || (_WIN32_WCE)
/*! \brief Windows systems use backslashes in file paths */
#define SLASH "\\"
#else
/*! \brief Most sane systems use forward-slashes */
#define SLASH "/"
#endif

/*! \brief Function to assemble a full path from a path and filename */
extern "C" DECLSPEC char * SDLCALL SprMakePath(const char *filename, const char *path, char *buf);

/*! \brief Class to handle loading all necessary data for a sprite  */
/**
 *  I began to realize that, with images, frames, bitmasks, layers,
 *  collision masks, positions, etc. etc. etc. all in one sprite,
 *  that was a heck of a lot of data to squeeze through a constructor,
 *  and quite tricky to get all in one place at once..
 *
 *  so I broke the sprite's initialization procedure down into a whole
 *  bunch of Init_Something() functions, and created this class
 *  to handle the tricky task of getting all the data ready for the
 *  sprite itself. 
 *
 *  It works by loading a sprite configuration file, which contains
 *  information on drawing layers, animation frames, collision masks and
 *  bitmasks, image names, and so forth.  The file is parsed, the data
 *  loaded, and once all that is done you can create as many identical
 *  sprites as you want from it.  This also allows you to create dozens
 *  of sprites of whatever kind you want, sharing the exact same data,
 *  cutting down on data duplication quite a bit.
 */
class DECLSPEC SpriteLoader
{
public:
  SpriteLoader();
  /*! \brief Loads a file specifying a sprite */
  SpriteLoader(const char *fname, const char *path="");
  /*! \brief Attempts to parse a .spr file and load the specified data  */
  bool Load(const char *fname, const char *path="");

  /*! \brief Tells if the load was successful */  
  bool Good(); 
  
  /*! \brief Creates or converts a sprite to the specs in the file */
  Sprite *GetSprite(Sprite *sprite=NULL);

  /*! \brief Returns # of y frames */
  inline int YFrames()              const { return(yframes);  }
  /*! \brief Returns # of x frames */  
  inline int XFrames()              const { return(xframes);  }
  /*! \brief Returns red transparency color   */
  inline Uint8 R()                  const { return(r);        }
  /*! \brief Returns blue transparency color  */
  inline Uint8 G()                  const { return(g);        }
  /*! \brief Returns green transparency color */
  inline Uint8 B()                  const { return(b);        }
  /*! \brief Returns bitmasks                         */
  inline bitmask **Bmask()          const { return(bmask);    }
  /*! \brief Returns the loaded image */
  inline SDL_Surface *GetSurface()  const { return(image);    }
  /*! \brief Returns the parsed collision layer mask  */
  inline ColType Layer()            const { return(colmask);  }
  /*! \brief Returns the parsed collision index mask  */
  inline ColType Index()            const { return(colindex); }
  /*! \brief Returns drawing layer  */
  inline int DrawLayer()            const { return(drawlayer);}

  /*! \brief Destructor, frees all loaded data */
  ~SpriteLoader();

protected:
  /* \brief Processes a single line from sprite-config file */
  bool ProcessLine(const char *buf);
  /*! \brief  Frees all data and returns variables to blank state */
  void Reset();  

  /*! \brief Collison bitmasks */
  bitmask    **bmask;
  /*! \brief Image */
  SDL_Surface *image;
  /*! \brief Transparency color */
  int r,g,b;
  /*! \brief Drawing layer */
  int drawlayer;
  /*! \brief Collision index mask. Uint32 so sscanf can use it */
  Uint32 colindex;
  /*! \brief Collision layer mask. Uint32 so sscanf can use it */
  Uint32 colmask;
  /*! \brief Status - true=successful load, false=error */
  bool good;
  /*! \brief true means create collision bitmask from image */
  bool loadbitmask;
  /*! \brief String containing image file name  */
  char imagename[300];
  /*! \brief Integers specifying number of rows and columns in image  */
  int xframes,yframes;
  /*! \brief Fixed point integer specifying frame rate */
  Sfxp framerate;
};

#include <SDL/close_code.h>

#endif
