/***************************************************************************
                          spritemanager.cpp  -  description
                             -------------------
    begin                : Fri Mar 7 2003
    copyright            : (C) 2003 by Tyler Montbriand
    email                : tsm@accesscomm.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <string.h>
#include <stdlib.h>
#include "spritemanager.h"
#include "ticktimer.h"
#include "spriteloader.h"

/**
 *  Unless this function returns 1, the SDL_USEREVENT is passed along
 *  to every sprite in the list.
 */
int SpriteManager::Event(int code, void *data1, void *data2)
{ return(0); }

/**
 *  SDL events that the event loop doesn't know how to handle
 *  get passed along to here.
 */
void SpriteManager::Handler(SDL_Event *event)
{ }

/**
 *  Game event loop calls this when it's not busy, can overload
 *  this to put in periodic checks and such
 */
void SpriteManager::Idle()
{ }


SpriteManager::SpriteManager(Uint16 w, Uint16 h, Uint8 depth, Uint32 vflags)
{
  flag=SM_TOTALREDRAW|SM_OWNBG;  

  lock=SDL_CreateMutex();

  if(!SDL_WasInit(SDL_INIT_VIDEO))
    if(SDL_InitSubSystem(SDL_INIT_VIDEO)<0)
    {
      fprintf(stderr,"Warning, couldn't initialize SDL video\n");
    }

  if(vflags&SDL_DOUBLEBUF)
  {
    fprintf(stderr,"SDL_DOUBLEBUF flag detected, using fullscreen flips\n");
    flag|=SM_DOUBLEBUF;
  }

  screen=SDL_SetVideoMode(w,h,depth,SDL_ANYFORMAT|vflags);

  if(screen==NULL)
  {
    fprintf(stderr,"SpriteManager::SpriteManager(%d,%d,%d,%d)\n",w,h,depth,vflags);
    fprintf(stderr,"Warning - can't set video mode!\n");
  }

  bg=SDL_DisplayFormat(screen);

  if(bg==NULL)
  {
    fprintf(stderr,"Error - can't create background!\n");
  }
}

void SpriteManager::DrawOnBG(SDL_Surface *src, SDL_Rect *srcrect, SDL_Rect *destrect)
{
  Lock();

  SDL_BlitSurface(src,srcrect,bg,destrect);

  flag|=SM_TOTALREDRAW;

  Unlock();
}

void SpriteManager::SendEvent(int code, void *data1, void *data2)
{
  Sprite *spr=sprites.Head();

  while(spr!=NULL)
  {
    spr->Event(code,data1,data2);
    spr=spr->next;
  }
}

/**
 *  Runs the game event loop until an SDL_QUIT message is
 *  recieved.  Calls nearly every other member while
 *  it's running, including the virtual Idle(), Event(),
 *  and Handler() functions that you can overload to do
 *  whatever you please.
 */
void SpriteManager::Run(Sint32 frametick)
{
  int ticks;
  TickTimer timer;
  flag|=SM_RUNNING;

  timer.Reset();
  /* Run the event loop until SDL_QUIT or enter/return keypress */
  while(flag&SM_RUNNING)
  {
    SDL_Event event;

    while(SDL_PollEvent(&event))/* SDL event loop */
    switch(event.type)
    {
    case SDL_QUIT:  /* Break the game loop on SDL_QUIT  */
      flag&=~SM_RUNNING;
      break;

    case SDL_USEREVENT:
      /* Pass SDL_USEREVENTs along to the virtual Event function  */
      if(!Event(event.user.code,event.user.data1,event.user.data2))
      {
        /* If Event() returns zero, pass the event to every sprite  */
        SendEvent(event.user.code, event.user.data1, event.user.data2);
      }
      break;

    default:  /* Pass any other events onto the virtual handler function */
      Handler(&event);
      break;
    }

    ticks=timer;

    /* If the screen needs updating */
    if(ticks>=frametick)
    {
      Draw();
      Update();

      if(!Paused()) /* Don't call Move() while paused */
      {
        /* Move in smaller increments if there's been a huge delay */
        while(ticks>MAX_FRAMETICK_GRANULARITY)
        {
          flag|=SM_TOTALREDRAW;
          ticks-=MAX_FRAMETICK_GRANULARITY;
          timer-=MAX_FRAMETICK_GRANULARITY;
          Move(MAX_FRAMETICK_GRANULARITY);
        }

        if(ticks>=frametick)
        {
          int n=ticks/frametick;
          Move(n*frametick);
          timer-=(n*frametick);
        }
      }

      /* Reset time count */
//      timer.Reset();
    }
    else /* Screen doesn't need updating */
    {
      int d=frametick-ticks;
      if(d>0) SDL_Delay(1);
        SDL_Delay(d); /* give time to other processes */
      Idle();         /* Call the virtual Idle handler  */
    }
  }
}

SpriteManager::SpriteManager(Uint16 flags,SDL_Surface *screenin,
                             const char *bgnd)
 :rects(MAX_RECTS)
{
  flag=flags|SM_TOTALREDRAW|SM_OWNBG;

  lock=SDL_CreateMutex();

  if(screenin==NULL)
  {
    screen=SDL_GetVideoSurface();
    if(screen==NULL)
    {
      fprintf(stderr,"Couldn't get video surface!\n");
	  exit(1);
    }

    if(bgnd==NULL)
    {
      bg=SDL_DisplayFormat(screen);
    }
    else
    {
      bg=LoadConvertBMP(bgnd);
    }
  }
  else
  {
    screen=screenin;

    if(bgnd==NULL)
    {
      bg=SDL_DisplayFormat(screen);
    }
    else
    {
      bg=LoadConvertBMP(bgnd);
    }
  }
}

bool SpriteManager::AddSprite(Sprite *spr)
{
  bool retval;

  Lock();
  retval=sprites.Insert(spr);
  Unlock();

  return(retval);
}

/**
 *  This doesn't destroy the sprite, just removes it from the list.
 */
bool SpriteManager::RemoveSprite(Sprite *spr)
{
  bool retval;

  Lock();
  retval=sprites.Extract(spr);
  Unlock();

  return(retval);
}

/**
 *  Draws the new frame, but doesn't update it.
 */
void SpriteManager::Draw()
{
  /* Redraw whole screen */
  if(flag&(SM_TOTALREDRAW|SM_DOUBLEBUF))
  {
    flag&=~SM_TOTALREDRAW;      /* Turn off total-redraw flag */
    SimpleBlit();               /* Draw */
  }
  else  /* Redraw only portion covered by sprites */
  {
    FastBlit();
  }
}

/**
 *  Handles sprite movement, removal, deletion, and collision.
 */
void SpriteManager::Move(Sint32 dtick)
{
  Sprite *spr=NULL;

  Lock(); /* Lock the mutex to prevent simultaneous access  */

  spr=sprites.Head();
  while(spr!=NULL)
  {
    if(spr->Expired())
    {
      Sprite *temp=spr->next;   /* Preserve pointer to next sprite in list */

      if(sprites.Extract(spr))  /* Remove sprite from list    */
      {
        if(spr->Deleteable())   /* Check if can delete sprite */
        {
          delete spr;           /* Delete the sprite          */
        }
        
        spr=temp; /* Restore pointer to next sprite */        
      }
      else
      {
        fprintf(stderr,"Error - can't remove expired sprite!\n");
        exit(1);
      }
      
      continue; /* Skip back up to top of while loop  */
    }

    if(spr->New())
    { /*  If the sprite's new, clear the Wait flag
          so it will be drawn next frame.           */
      spr->ClearFlags(SPRITE_WAIT);
      /*  Set the MOVED flag so it will be checked for
          collisions against all sprites at least once. */
      spr->SetFlags(SPRITE_MOVED);
    }
    
    spr->Update(dtick);     /* Call the virtual update function */

    spr=spr->next;  /* Proceed to next sprite in list */
  }

  /* Optimize me - Don't test unmoving sprites against unmoving sprites */
  spr=sprites.Head();
  while(spr!=NULL)  /* Loop through every sprite in list  */
  {
    Sprite *spr2=spr->next; /* Get next sprite in list  */

    if(spr->Moved())  /* Test against moving and unmoving sprites */
    {
      while(spr2!=NULL) /* Loop through portion of list after spr */
      {
        if(spr->CanCollide(spr2)) /* If the sprites collide */
        {
          /* Call their virtual collision handler functions */
          spr->OnCollide(  spr2,spr2->GetColIndex() );
          spr2->OnCollide( spr , spr->GetColIndex() );
        }

        spr2=spr2->next;  /* Advance to next sprite in inner loop */
      }
    }
    else /* Only test against moving sprites */
    {
      while(spr2!=NULL) /* Loop through portion of list after spr */
      {
        if(spr2->Moved())
          if(spr->CanCollide(spr2)) /* If the sprites collide */
          {
            /* Call their virtual collision handler functions */
            spr->OnCollide(  spr2,spr2->GetColIndex() );
            spr2->OnCollide( spr , spr->GetColIndex() );
          }

        spr2=spr2->next;  /* Advance to next sprite in inner loop */
      }      
    }

    spr=spr->next;      /* Advance to next sprite in outer loop */
  }

  Unlock(); /*  Unlock the mutex to allow access to the manager */
}

/**
 *  Frees the background and mutex.
 */
SpriteManager::~SpriteManager()
{
  Lock(); /* Lock mutex to prevent simultaneous access  */

  /* If the background is freeable */
  if((flag&SM_OWNBG)&&(bg!=NULL))
  {
    SDL_FreeSurface(bg);  /* Free the background surface */
  }
  
  SDL_DestroyMutex(lock); /* Destroy the mutex itself */
}
