/***************************************************************************
                          spritemanager.h  -  description
                             -------------------
    begin                : Fri Mar 7 2003
    copyright            : (C) 2003 by Tyler Montbriand
    email                : tsm@accesscomm.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef __SPRITEMANAGER_H__
#define __SPRITEMANAGER_H__

#include "RotateImage.h"
#include "Fxp.h"
#include "rect.h"
#include "rectlist.h"
#include "sprite.h"
#include "spritelist.h"
#include "ticktimer.h"
#include "spriteloader.h"

#include <SDL/begin_code.h>

/*! \brief Maximum amount of time that can be passed to sprites */
#define MAX_FRAMETICK_GRANULARITY 200
/*! \brief Maximum number of rectangles in draw-stack           */
#define MAX_RECTS       512
/*! \brief No flags */
#define SM_NONE         0x0000
/*! \brief SpriteManager is running */
#define SM_RUNNING      0x0001
/*! \brief SpriteManager needs to totally redraw screen */
#define SM_TOTALREDRAW  0x0002
/*! \brief SpriteManager is paused  */
#define SM_PAUSED       0x0004
/*! \brief SpriteManager owns the background surface  */
#define SM_OWNBG        0x0008
/*! \brief Spritemanager must do full screen flips */
#define SM_DOUBLEBUF    0x0010

/*! \brief Class to manage movement, collision, and drawing of sprites */
class DECLSPEC SpriteManager
{
public:
  /*! \brief Constructor that sets screen mode  */
  SpriteManager(Uint16 w, Uint16 h, Uint8 depth, Uint32 vflags);
  /*! \brief Constructor - all parameters optional */
  SpriteManager(Uint16 flags=0,SDL_Surface *screenin=NULL, const char *bgnd=NULL);
  /*! \brief Attempts to add a sprite to the list */
  bool AddSprite(Sprite *spr);
  /*! \brief Attempts to remove a sprite from the list  */
  bool RemoveSprite(Sprite *spr);
  /*! \brief Draws the sprites on the screen  */
  void Draw();
  /*! \brief Calls the Update() member of every sprite */
  void Move(Sint32 dtick);
  /*! \brief Runs the game event loop until SDL_QUIT  */
  void Run(Sint32 frametick);
  /*! \brief Calls the Event() member of every sprite with this data  */
  void SendEvent(int code, void *data1=NULL, void *data2=NULL);
  /*! \brief Blits specified portion of surface onto background image */
  void DrawOnBG(SDL_Surface *src, SDL_Rect *srcrect, SDL_Rect *destrect);

  /*! \brief Handler function for SDL_USEREVENTs  */
  virtual int Event(int code, void *data1, void *data2);
  /*! \brief Function called when event loop has time to kill */
  virtual void Idle();
  /*! \brief Function called for unhandled SDL events */
  virtual void Handler(SDL_Event *event);

  /*! \brief Destructor, frees all data */
  virtual ~SpriteManager();

  /*! \brief Sets SM_TOTALREDRAW flag */
  inline void Redraw()
  {
    flag|=SM_TOTALREDRAW;
  }

  /*! \brief Updates relevant portions of the screen */
  inline void Update()
  {
    if(flag&SM_DOUBLEBUF) /* SM_DOUBLEBUF forces use of SDL_Flip() */
      SDL_Flip(screen);
    else /* Otherwise, do partial screen update */
      rects.Update(screen);
    
    rects.Clear();
  }

  /*! Returns nonzero if paused, zero if not  */
  inline int Paused()
  {
    return(flag&SM_PAUSED);
  }

  /*! \brief Sets the SM_PAUSED flag  */
  inline void Pause()
  {
    Lock();
    flag|=SM_PAUSED;
    Unlock();
  }

  /*! \brief Clears the SM_PAUSED flag  */
  inline void Unpause()
  {
    Lock();
    flag&=~SM_PAUSED;
    Unlock();
  }

  /*! \brief Returns the background surface and locks the SpriteManager */
  SDL_Surface *CheckOutBG()
  {
    Lock();
    return(bg);
  }

  /*! \brief Checks the surface back in and unlocks the SpriteManager */
  bool CheckInBG(SDL_Surface *bg_in)
  {
    if(bg_in==bg)
    {
      Unlock();
      return(true);
    }
    else
    {
      return(false);
    }
  }

  /*! \brief Attempts to add the specified rect to the screen update list */
  inline bool AddUpdateRect(SDL_Rect &r)
  {
    bool retval;

    Lock();
    retval=rects.Push(Rect(r));
    Unlock();

    return(retval);
  }

protected:

  /*! \brief Locks the mutex, to prevent simultaneous access  */
  inline void Lock()
  {
    SDL_mutexP(lock);
  }
  /*! \brief Unlocks the mutex, leaving the manager free for use  */
  inline void Unlock()
  {
    SDL_mutexV(lock);
  }

  /*! \brief Updates the entire screen  */
  inline void SimpleBlit()
  {
    Sprite *spr=NULL;
    rects.Clear();
    rects.Push(Rect(0,0,bg->w,bg->h));

    SDL_BlitSurface(bg,NULL,screen,NULL);

    spr=sprites.Head();

    while(spr!=NULL)
    {
      if(!spr->Expired()) spr->Draw(screen);

      spr=spr->next;
    }
  }

  /*! \brief Redraws only sprites and the areas under them  */
  inline void FastBlit()
  {
    Sprite *spr=NULL;

    for(spr=sprites.Head(); spr!=NULL; spr=spr->next)
    {
      Rect rect,prev;
      rect=spr->CurRect();
      prev=spr->PrevRect();

      if(rect.Merge(prev))
      {
        SDL_BlitSurface(bg,&rect,screen,&rect);
        rects.Push(rect);
      }
      else
      {
        SDL_BlitSurface(bg,&rect,screen,&rect);
        SDL_BlitSurface(bg,&prev,screen,&prev);
        rects.Push(rect);
        rects.Push(prev);
      }
    }

    for(spr=sprites.Head(); spr!=NULL; spr=spr->next)
    {
      if(spr->Expired())  continue;
      spr->Draw(screen);
    }
  }

  /*! \brief Status flags */
  Uint16       flag;
  /*! \brief Mutex;  locks during operations to make it threadsafe */
  SDL_mutex   *lock;
  /*!\brief The surface that gets drawn to */
  SDL_Surface *screen;
  /*! \brief The background image */
  SDL_Surface *bg;
  /*! \brief Linked list of sprites */  
  SpriteList   sprites;
  /*! \brief Big stack of rectangles */
  RectList     rects;
};

#include <SDL/close_code.h>

#endif/*__SPRITEMANAGER_H__*/
