/***************************************************************************
                  ticktimer.h  -  Simple ms-based timing class
                             -------------------
    begin                : Thu Mar 6 2003
    copyright            : (C) 2003 by Tyler Montbriand
    email                : tsm@accesscomm.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TICKTIMER_H
#define TICKTIMER_H

#include <SDL/SDL_timer.h>

/**
  *@author Tyler Montbriand
  */

/*! \brief Simple stopwatch class, based on SDL_GetTicks() */
class TickTimer
{
public:
  /*! \brief Constructor. */
  inline TickTimer()
  {
    Reset();
  }

  /*! \brief Set the timer back to zero. */  
  inline void Reset()
  {
    lasttick=GetCurTime();
  }

  inline Sint32 Left(Sint32 time)
  {
    int t=Ticks();
    /* Can't wait backwards */
    if(time<0)      return(0);
    else if(t>time) return(0);
    else            return(t-time);
  }

  /*! \brief Returns the time since last reset */
  inline Sint32 Ticks()
  {
    return(GetCurTime()-lasttick);
  }

  /*! \brief Increments output time */
  inline void operator+=(Sint32 ticks)
  {
    lasttick-=ticks;
  }

  /*! \brief Decrements output time */
  inline void operator-=(Sint32 ticks)
  {
    lasttick+=ticks;
  }

  /*! \brief Makes it act like an integer representing time since last reset */
  inline operator Sint32()
  { return(Ticks());  }
  
protected:
  /*! \brief Overloadable function to get time reference */
  virtual Sint32 GetCurTime()
  { return(SDL_GetTicks());  }
  
  /*! \brief The starting point; reference time */
  Sint32 lasttick;
};

/*! \brief Simple stopwatch class based on a variable time reference */
class ManualTimer:public TickTimer
{
public:
  ManualTimer():TickTimer()
  {
    lasttick=0;
    curtime=0;        
  }

  /*! \brief Advances the time reference forward */  
  inline void Advance(Sint32 tick)
  { if(tick>0)  curtime+=tick;  }
  
protected:
  /*! \brief Function overloaded to return variable */
  virtual Sint32 GetCurTime()
  { return curtime; }
  /*! \brief Fixed time reference */
  Sint32 curtime;  
};

class DebugTimer
{
public:
  /*! \brief Constructor. */
  inline DebugTimer(int step=100)
  {
    lasttick=step;
  }

  /*! \brief Set the timer back to zero. */
  inline void Reset()
  {
  }

  inline Sint32 Left(Sint32 time)
  {
    int t=Ticks();
    /* Can't wait backwards */
    if(time<0)      return(0);
    else if(t>time) return(0);
    else            return(t-time);
  }

  /*! \brief Returns the time since last reset */
  inline Sint32 Ticks()
  {
    return(lasttick);
  }

  /*! \brief Makes it act like an integer representing time since last reset */
  inline operator Sint32()
  { return(Ticks());  }

protected:

  /*! \brief The starting point; reference time */
  Sint32 lasttick;
};

#endif
